"""Testing DKI"""

import random
import warnings

import numpy as np
from numpy.testing import (
    assert_almost_equal,
    assert_array_almost_equal,
    assert_array_equal,
    assert_raises,
)

from dipy.core.geometry import sphere2cart
from dipy.core.gradients import gradient_table
from dipy.core.sphere import Sphere
from dipy.data import default_sphere, get_fnames
from dipy.io.gradients import read_bvals_bvecs
import dipy.reconst.dki as dki
from dipy.reconst.dki import (
    MIN_POSITIVE_SIGNAL,
    _positive_evals,
    axial_kurtosis,
    carlson_rd,
    carlson_rf,
    kurtosis_fractional_anisotropy,
    lower_triangular,
    ls_fit_dki,
    mean_kurtosis,
    mean_kurtosis_tensor,
    radial_kurtosis,
    radial_tensor_kurtosis,
)
import dipy.reconst.dti as dti
from dipy.reconst.dti import decompose_tensor, from_lower_triangular
from dipy.reconst.utils import dki_design_matrix
from dipy.reconst.weights_method import (
    weights_method_nlls_m_est,
    weights_method_wls_m_est,
)
from dipy.sims.voxel import multi_tensor_dki
from dipy.testing import check_for_warnings
from dipy.utils.optpkg import optional_package
from dipy.utils.tripwire import TripWireError

cvxpy, have_cvxpy, _ = optional_package("cvxpy", min_version="1.4.1")

gtab, gtab_2s, crossing_ref, signal_cross = None, None, None, None
multi_params, Kref_sphere, DWI = None, None, None
mevals_cross, angles_cross, frac_cross, kt_cross = None, None, None, None
dt_sph, evals_sph, kt_sph, params_sph = None, None, None, None


def setup_module():
    global gtab, gtab_2s, crossing_ref, signal_cross
    global multi_params, Kref_sphere, DWI, S0
    global mevals_cross, angles_cross, frac_cross, kt_cross
    global dt_sph, evals_sph, kt_sph, params_sph

    fimg, fbvals, fbvecs = get_fnames(name="small_64D")
    bvals, bvecs = read_bvals_bvecs(fbvals, fbvecs)
    gtab = gradient_table(bvals, bvecs=bvecs)

    # 2 shells for techniques that requires multishell data
    bvals_2s = np.concatenate((bvals, bvals * 2), axis=0)
    bvecs_2s = np.concatenate((bvecs, bvecs), axis=0)
    gtab_2s = gradient_table(bvals_2s, bvecs=bvecs_2s)

    # Simulation 1. signals of two crossing fibers are simulated
    mevals_cross = np.array(
        [
            [0.00099, 0, 0],
            [0.00226, 0.00087, 0.00087],
            [0.00099, 0, 0],
            [0.00226, 0.00087, 0.00087],
        ]
    )
    angles_cross = [(80, 10), (80, 10), (20, 30), (20, 30)]
    fie = 0.49
    frac_cross = [fie * 50, (1 - fie) * 50, fie * 50, (1 - fie) * 50]
    S0 = 100
    # Noise free simulates
    signal_cross, dt_cross, kt_cross = multi_tensor_dki(
        gtab_2s,
        mevals_cross,
        S0=S0,
        angles=angles_cross,
        fractions=frac_cross,
        snr=None,
    )
    evals_cross, evecs_cross = decompose_tensor(from_lower_triangular(dt_cross))
    crossing_ref = np.concatenate(
        (evals_cross, evecs_cross[0], evecs_cross[1], evecs_cross[2], kt_cross), axis=0
    )

    # Simulation 2. Spherical kurtosis tensor.- for white matter, this can be a
    # biological implausible scenario, however this simulation is useful for
    # testing the estimation of directional apparent kurtosis and the mean
    # kurtosis, since its directional and mean kurtosis ground truth are a
    # constant which can be calculated easily mathematically.
    Di = 0.00099
    De = 0.00226
    mevals_sph = np.array([[Di, Di, Di], [De, De, De]])
    frac_sph = [50, 50]
    signal_sph, dt_sph, kt_sph = multi_tensor_dki(
        gtab_2s, mevals_sph, S0=S0, fractions=frac_sph, snr=None
    )
    evals_sph, evecs_sph = decompose_tensor(from_lower_triangular(dt_sph))
    params_sph = np.concatenate(
        (evals_sph, evecs_sph[0], evecs_sph[1], evecs_sph[2], kt_sph), axis=0
    )
    # Compute ground truth - since KT is spherical, apparent kurtosis
    # coefficient for all gradient directions and mean kurtosis have to be
    # equal to Kref_sph.
    f = 0.5
    Dg = f * Di + (1 - f) * De
    Kref_sphere = 3 * f * (1 - f) * ((Di - De) / Dg) ** 2

    # Simulation 3. Multi-voxel simulations - dataset of four voxels is
    # simulated. Since the objective of this simulation is to see if
    # procedures are able to work with multi-dimensional data all voxels
    # contains the same crossing signal produced in simulation 1.

    DWI = np.zeros((2, 2, 1, len(gtab_2s.bvals)))
    DWI[0, 0, 0] = DWI[0, 1, 0] = DWI[1, 0, 0] = DWI[1, 1, 0] = signal_cross
    multi_params = np.zeros((2, 2, 1, 27))
    multi_params[0, 0, 0] = multi_params[0, 1, 0] = crossing_ref
    multi_params[1, 0, 0] = multi_params[1, 1, 0] = crossing_ref


def teardown_module():
    global gtab, gtab_2s, crossing_ref, signal_cross
    global multi_params, Kref_sphere, DWI
    global mevals_cross, angles_cross, frac_cross, kt_cross
    global dt_sph, evals_sph, kt_sph, params_sph
    gtab, gtab_2s, crossing_ref, signal_cross = None, None, None, None
    multi_params, Kref_sphere, DWI = None, None, None
    mevals_cross, angles_cross, frac_cross, kt_cross = None, None, None, None
    dt_sph, evals_sph, kt_sph, params_sph = None, None, None, None


def test_positive_evals():
    # Tested evals
    L1 = np.array([[1e-3, 1e-3, 2e-3], [0, 1e-3, 0]])
    L2 = np.array([[3e-3, 0, 2e-3], [1e-3, 1e-3, 0]])
    L3 = np.array([[4e-3, 1e-4, 0], [0, 1e-3, 0]])
    # only the first voxels have all eigenvalues larger than zero, thus:
    expected_ind = np.array([[True, False, False], [False, True, False]], dtype=bool)
    # test function _positive_evals
    ind = _positive_evals(L1, L2, L3)
    assert_array_equal(ind, expected_ind)


def test_split_dki_param():
    dkiM = dki.DiffusionKurtosisModel(gtab_2s, fit_method="OLS")
    dkiF = dkiM.fit(DWI)
    evals, evecs, kt = dki.split_dki_param(dkiF.model_params)

    assert_array_almost_equal(evals, dkiF.evals)
    assert_array_almost_equal(evecs, dkiF.evecs)
    assert_array_almost_equal(kt, dkiF.kt)


def test_dki_fits():
    """DKI fits are tested on noise free crossing fiber simulates"""

    mask_signal_cross = np.ones_like(signal_cross)
    mask_signal_cross[..., ::2] = 0
    mask_signal_cross[0] = 1

    # OLS fitting
    dkiM = dki.DiffusionKurtosisModel(gtab_2s, fit_method="OLS")
    dkiF = dkiM.fit(signal_cross)

    assert_array_almost_equal(dkiF.model_params, crossing_ref)
    dkiF = dkiM.fit(signal_cross, mask=mask_signal_cross)
    assert_array_almost_equal(dkiF.model_params, crossing_ref)

    # OLS fitting with return leverages
    dkiM = dki.DiffusionKurtosisModel(gtab_2s, fit_method="OLS", return_leverages=True)
    dkiF = dkiM.fit(signal_cross)
    leverages = dkiM.extra["leverages"]
    assert_almost_equal(leverages.sum(), np.array([22.0]))

    # WLS fitting
    dki_wlsM = dki.DiffusionKurtosisModel(gtab_2s, fit_method="WLS")
    dki_wlsF = dki_wlsM.fit(signal_cross)

    assert_array_almost_equal(dki_wlsF.model_params, crossing_ref)

    # Test wls given S^2 weights argument, matches default wls
    design_matrix = dki_design_matrix(gtab_2s)
    inverse_design_matrix = np.linalg.pinv(design_matrix)
    YN = signal_cross + 10 * np.random.normal(size=signal_cross.shape)  # error
    YN[YN < MIN_POSITIVE_SIGNAL] = MIN_POSITIVE_SIGNAL
    # wls calculation
    D_w, _ = ls_fit_dki(
        design_matrix,
        YN,
        inverse_design_matrix,
        return_S0_hat=False,
        return_lower_triangular=True,
        weights=True,
    )
    # wls calculation, by calculating S^2 from OLS fit, then passing weights
    D_o, _ = ls_fit_dki(
        design_matrix,
        YN,
        inverse_design_matrix,
        return_S0_hat=False,
        return_lower_triangular=True,
        weights=False,
    )
    pred_s = np.exp(np.dot(design_matrix, D_o.T)).T
    D_W, _ = ls_fit_dki(
        design_matrix,
        YN,
        inverse_design_matrix,
        return_S0_hat=False,
        return_lower_triangular=True,
        weights=pred_s**2,
    )  # S^2 weights
    assert_almost_equal(D_w, D_W)
    # wls calculation, but passing incorrect weights
    D_W, _ = ls_fit_dki(
        design_matrix,
        YN,
        inverse_design_matrix,
        return_S0_hat=False,
        return_lower_triangular=True,
        weights=pred_s**1,
    )
    assert_raises(AssertionError, assert_array_equal, D_w, D_W)
    # Test that wls implementation is correct, by comparison with result here
    W = np.diag(pred_s.squeeze() ** 2)
    AT_W = np.dot(design_matrix.T, W)
    inv_AT_W_A = np.linalg.pinv(np.dot(AT_W, design_matrix))
    AT_W_LS = np.dot(AT_W, np.log(YN).squeeze())
    result = np.dot(inv_AT_W_A, AT_W_LS)
    assert_almost_equal(D_w.squeeze(), result)

    if have_cvxpy:
        # CLS fitting
        dki_clsM = dki.DiffusionKurtosisModel(
            gtab_2s, fit_method="CLS", cvxpy_solver=cvxpy.CLARABEL
        )
        dki_clsF = dki_clsM.fit(signal_cross)

        assert_array_almost_equal(dki_clsF.model_params, crossing_ref)

        # CWLS fitting
        dki_cwlsM = dki.DiffusionKurtosisModel(
            gtab_2s, fit_method="CWLS", cvxpy_solver=cvxpy.CLARABEL
        )
        dki_cwlsF = dki_cwlsM.fit(signal_cross)

        assert_array_almost_equal(dki_cwlsF.model_params, crossing_ref)
    else:
        assert_raises(
            TripWireError, dki.DiffusionKurtosisModel, gtab_2s, fit_method="CLS"
        )
        assert_raises(
            TripWireError, dki.DiffusionKurtosisModel, gtab_2s, fit_method="CWLS"
        )

    # NLS fitting
    if have_cvxpy:
        dki_nlsM = dki.DiffusionKurtosisModel(gtab_2s, fit_method="NLS")
        dki_nlsF = dki_nlsM.fit(signal_cross)
        assert_array_almost_equal(dki_nlsF.model_params, crossing_ref)
        dki_nlsF = dki_nlsM.fit(signal_cross, mask=mask_signal_cross)
        assert_array_almost_equal(dki_nlsF.model_params, crossing_ref)
    else:
        assert_raises(
            TripWireError, dki.DiffusionKurtosisModel, gtab_2s, fit_method="NLS"
        )

    # Restore fitting
    dki_rtM = dki.DiffusionKurtosisModel(gtab_2s, fit_method="RESTORE", sigma=2)
    dki_rtF = dki_rtM.fit(signal_cross)
    assert_array_almost_equal(dki_rtF.model_params, crossing_ref)

    # Test single voxel return_S0
    if have_cvxpy:
        dki_nlsM = dki.DiffusionKurtosisModel(
            gtab_2s, fit_method="NLS", return_S0_hat=True, cvxpy_solver=cvxpy.CLARABEL
        )
        dki_nlsF = dki_nlsM.fit(signal_cross)
        assert_array_almost_equal(dki_nlsF.model_S0, S0)
    else:
        assert_raises(
            TripWireError, dki.DiffusionKurtosisModel, gtab_2s, fit_method="NLS"
        )

    # testing multi-voxels
    mask_signal_multi = np.ones_like(DWI[..., 0])
    mask_signal_multi[1, 1, ...] = 0

    dkiF_multi = dkiM.fit(DWI)
    assert_array_almost_equal(dkiF_multi.model_params, multi_params)

    dkiF_multi = dki_wlsM.fit(DWI)
    assert_array_almost_equal(dkiF_multi.model_params, multi_params)

    dkiF_multi = dki_rtM.fit(DWI)
    assert_array_almost_equal(dkiF_multi.model_params, multi_params)

    if have_cvxpy:
        dkiF_multi = dki_nlsM.fit(DWI)
        assert_array_almost_equal(dkiF_multi.model_params, multi_params)
        dkiF_multi = dki_nlsM.fit(DWI, mask=mask_signal_multi)
        masked_multi_params = multi_params.copy()
        masked_multi_params[1, 1, ...] = 0
        assert_array_almost_equal(dkiF_multi.model_params, masked_multi_params)

    # testing return of S0
    if have_cvxpy:
        tested_methods = ["NLLS", "WLLS", "CLS", "CWLS"]
    else:
        tested_methods = ["NLLS", "WLLS"]

    for fit_method in tested_methods:
        kwargs = {}
        if fit_method in ["CLS", "CWLS"]:
            kwargs = {"cvxpy_solver": cvxpy.CLARABEL}
        dki_S0M = dki.DiffusionKurtosisModel(
            gtab_2s, fit_method=fit_method, return_S0_hat=True, **kwargs
        )
        dki_S0F = dki_S0M.fit(DWI)
        dki_S0F_S0 = dki_S0F.model_S0

        assert_array_almost_equal(dki_S0F_S0, np.full(dki_S0F_S0.shape, S0))

    for fit_method in tested_methods:
        kwargs = {}
        if fit_method in ["CLS", "CWLS"]:
            kwargs = {"cvxpy_solver": cvxpy.CLARABEL}
        dki_S0M = dki.DiffusionKurtosisModel(
            gtab_2s, fit_method=fit_method, return_S0_hat=True, **kwargs
        )
        # FIXME: CLS and CWLS generally fail with noisy data - needs fixing
        # dki_S0F = dki_S0M.fit(DWI + np.random.normal(size=DWI.shape),\
        #                      mask=mask_test)
        dki_S0F = dki_S0M.fit(DWI)
        dki_S0F_S0 = dki_S0F.model_S0

        assert_array_almost_equal(dki_S0F_S0, np.full(dki_S0F_S0.shape, S0))

    # testing robust fitting by with WLS and CWLS (needs noisy data)
    # --------==============----------------------==================
    YN = DWI + np.random.normal(size=DWI.shape)
    if have_cvxpy:
        tested_methods = ["WLS"]
        # FIXME: CWLS generally failing with noisy data, needs fixing
        # tested_methods = ['WLS', 'CWLS']
    else:
        tested_methods = ["WLS"]
    for fit_method in tested_methods:
        kwargs = {}
        if fit_method == "CWLS":
            kwargs = {"cvxpy_solver": cvxpy.CLARABEL}
        wm = weights_method_wls_m_est  # weights method
        dki_M = dki.DiffusionKurtosisModel(
            gtab_2s,
            fit_method=fit_method,
            return_S0_hat=True,
            weights_method=wm,
            fit_type=fit_method,
            num_iter=4,
            **kwargs,
        )

        for mask in [None, np.ones(DWI.shape[0:-1])]:
            # Smoke test
            _ = dki_M.fit(YN, mask=mask)

        # test that a single iteration will raise an error
        dki_M = dki.DiffusionKurtosisModel(
            gtab_2s,
            fit_method=fit_method,
            return_S0_hat=True,
            weights_method=wm,
            fit_type=fit_method,
            num_iter=1,
            **kwargs,
        )

        assert_raises(ValueError, dki_M.fit, YN)

    # RWLS and RNLLS work based are dti fit functions
    tested_methods = ["RWLS", "RNLLS"]
    for fit_method in tested_methods:
        dki_M = dki.DiffusionKurtosisModel(
            gtab_2s, fit_method=fit_method, return_S0_hat=True
        )
        # Smoke test: check that the fit method runs without error
        _ = dki_M.fit(YN)
        _ = dki_M.fit(YN, mask=np.ones(DWI.shape[0:-1]))  # test with mask

        # try wrong mask (these are not 'multi_fit' methods)
        assert_raises(
            ValueError, dki_M.fit, YN, mask=np.tile(np.ones(DWI.shape[0:-1]), 2)
        )

    # use RWLS and RNLLS but via explicit call to IRLS
    tested_methods = ["WLS", "NLLS"]
    for fm in tested_methods:
        wm = weights_method_wls_m_est if fm == "WLS" else weights_method_nlls_m_est
        dki_M = dki.DiffusionKurtosisModel(
            gtab_2s,
            fit_method="IRLS",
            return_S0_hat=True,
            weights_method=wm,
            fit_type=fm,
        )  # num_iter default
        _ = dki_M.fit(YN)

    # test the weights methods fail if num_iter is too low
    tested_methods = ["WLS", "NLLS"]
    for fm in tested_methods:
        wm = weights_method_wls_m_est if fm == "WLS" else weights_method_nlls_m_est
        dki_M = dki.DiffusionKurtosisModel(
            gtab_2s,
            fit_method="IRLS",
            return_S0_hat=True,
            weights_method=wm,
            fit_type=fm,
            num_iter=2,
        )

        assert_raises(ValueError, dki_M.fit, YN)


def test_apparent_kurtosis_coef():
    """Apparent kurtosis coefficients are tested for a spherical kurtosis
    tensor"""

    sph = Sphere(xyz=gtab.bvecs[gtab.bvals > 0])
    AKC = dki.apparent_kurtosis_coef(params_sph, sph)

    # check all direction
    for d in range(len(gtab.bvecs[gtab.bvals > 0])):
        assert_array_almost_equal(AKC[d], Kref_sphere)


def test_dki_predict():
    dkiM = dki.DiffusionKurtosisModel(gtab_2s)
    pred = dkiM.predict(crossing_ref, S0=100)

    assert_array_almost_equal(pred, signal_cross)

    # just to check that it works with more than one voxel:
    pred_multi = dkiM.predict(multi_params, S0=100)
    assert_array_almost_equal(pred_multi, DWI)

    # Check that it works with more than one voxel, and with a different S0
    # in each voxel:
    pred_multi = dkiM.predict(multi_params, S0=100 * np.ones(pred_multi.shape[:3]))
    assert_array_almost_equal(pred_multi, DWI)

    # check the function predict of the DiffusionKurtosisFit object
    dkiF = dkiM.fit(DWI)
    pred_multi = dkiF.predict(gtab_2s, S0=100)
    assert_array_almost_equal(pred_multi, DWI)

    dkiF = dkiM.fit(pred_multi)
    pred_from_fit = dkiF.predict(dkiM.gtab, S0=100)
    assert_array_almost_equal(pred_from_fit, DWI)

    # Test the module function:
    pred = dki.dki_prediction(crossing_ref, gtab_2s, S0=100)
    assert_array_almost_equal(pred, signal_cross)

    # Test the module function with S0 volume:
    pred = dki.dki_prediction(
        multi_params, gtab_2s, S0=100 * np.ones(multi_params.shape[:3])
    )
    assert_array_almost_equal(pred, DWI)


def test_carlson_rf():
    # Define inputs that we know the outputs from:
    # Carlson, B.C., 1994. Numerical computation of real or complex
    # elliptic integrals. arXiv:math/9409227 [math.CA]

    # Real values (test in 2D format)
    x = np.array([[1.0, 0.5], [2.0, 2.0]])
    y = np.array([[2.0, 1.0], [3.0, 3.0]])
    z = np.array([[0.0, 0.0], [4.0, 4.0]])

    # Define reference outputs
    RF_ref = np.array(
        [[1.3110287771461, 1.8540746773014], [0.58408284167715, 0.58408284167715]]
    )

    # Compute integrals
    RF = carlson_rf(x, y, z)

    # Compare
    assert_array_almost_equal(RF, RF_ref)

    # Complex values
    x = np.array([1j, 1j - 1, 1j, 1j - 1])
    y = np.array([-1j, 1j, -1j, 1j])
    z = np.array([0.0, 0.0, 2, 1 - 1j])

    # Define reference outputs
    RF_ref = np.array(
        [
            1.8540746773014,
            0.79612586584234 - 1.2138566698365j,
            1.0441445654064,
            0.93912050218619 - 0.53296252018635j,
        ]
    )
    # Compute integrals
    RF = carlson_rf(x, y, z, errtol=3e-5)

    # Compare
    assert_array_almost_equal(RF, RF_ref)


def test_carlson_rd():
    # Define inputs that we know the outputs from:
    # Carlson, B.C., 1994. Numerical computation of real or complex
    # elliptic integrals. arXiv:math/9409227 [math.CA]

    # Real values
    x = np.array([0.0, 2.0])
    y = np.array([2.0, 3.0])
    z = np.array([1.0, 4.0])

    # Defene reference outputs
    RD_ref = np.array([1.7972103521034, 0.16510527294261])

    # Compute integrals
    RD = carlson_rd(x, y, z, errtol=1e-5)

    # Compare
    assert_array_almost_equal(RD, RD_ref)

    # Complex values (testing in 2D format)
    x = np.array([[1j, 0.0], [0.0, -2 - 1j]])
    y = np.array([[-1j, 1j], [1j - 1, -1j]])
    z = np.array([[2.0, -1j], [1j, -1 + 1j]])

    # Defene reference outputs
    RD_ref = np.array(
        [
            [0.65933854154220, 1.2708196271910 + 2.7811120159521j],
            [-1.8577235439239 - 0.96193450888839j, 1.8249027393704 - 1.2218475784827j],
        ]
    )
    # Compute integrals
    RD = carlson_rd(x, y, z, errtol=1e-5)

    # Compare
    assert_array_almost_equal(RD, RD_ref)


def test_Wrotate_single_fiber():
    # Rotate the kurtosis tensor of single fiber simulate to the diffusion
    # tensor diagonal and check that is equal to the kurtosis tensor of the
    # same single fiber simulated directly to the x-axis

    # Define single fiber simulate
    mevals = np.array([[0.00099, 0, 0], [0.00226, 0.00087, 0.00087]])
    fie = 0.49
    frac = [fie * 100, (1 - fie) * 100]

    # simulate single fiber not aligned to the x-axis
    theta = random.uniform(0, 180)
    phi = random.uniform(0, 320)
    angles = [(theta, phi), (theta, phi)]
    signal, dt, kt = multi_tensor_dki(
        gtab_2s, mevals, angles=angles, fractions=frac, snr=None
    )

    evals, evecs = decompose_tensor(from_lower_triangular(dt))

    kt_rotated = dki.Wrotate(kt, evecs)
    # Now coordinate system has the DT diagonal aligned to the x-axis

    # Reference simulation in which DT diagonal is directly aligned to the
    # x-axis
    angles = (90, 0), (90, 0)
    signal, dt_ref, kt_ref = multi_tensor_dki(
        gtab_2s, mevals, angles=angles, fractions=frac, snr=None
    )

    assert_array_almost_equal(kt_rotated, kt_ref)


def test_Wrotate_crossing_fibers():
    # Test 2 - simulate crossing fibers intersecting at 70 degrees.
    # In this case, diffusion tensor principal eigenvector will be aligned in
    # the middle of the crossing fibers. Thus, after rotating the kurtosis
    # tensor, this will be equal to a kurtosis tensor simulate of crossing
    # fibers both deviating 35 degrees from the x-axis. Moreover, we know that
    # crossing fibers will be aligned to the x-y plane, because the smaller
    # diffusion eigenvalue, perpendicular to both crossings fibers, will be
    # aligned to the z-axis.

    # Simulate the crossing fiber
    angles = [(90, 30), (90, 30), (20, 30), (20, 30)]
    fie = 0.49
    frac = [fie * 50, (1 - fie) * 50, fie * 50, (1 - fie) * 50]
    mevals = np.array(
        [
            [0.00099, 0, 0],
            [0.00226, 0.00087, 0.00087],
            [0.00099, 0, 0],
            [0.00226, 0.00087, 0.00087],
        ]
    )

    signal, dt, kt = multi_tensor_dki(
        gtab_2s, mevals, angles=angles, fractions=frac, snr=None
    )

    evals, evecs = decompose_tensor(from_lower_triangular(dt))

    kt_rotated = dki.Wrotate(kt, evecs)
    # Now coordinate system has diffusion tensor diagonal aligned to the x-axis

    # Simulate the reference kurtosis tensor
    angles = [(90, 35), (90, 35), (90, -35), (90, -35)]

    signal, dt, kt_ref = multi_tensor_dki(
        gtab_2s, mevals, angles=angles, fractions=frac, snr=None
    )

    # Compare rotated with the reference
    assert_array_almost_equal(kt_rotated, kt_ref)


def test_Wcons():
    # Construct the 4D kurtosis tensor manually from the crossing fiber kt
    # simulate
    Wfit = np.zeros([3, 3, 3, 3])

    # Wxxxx
    Wfit[0, 0, 0, 0] = kt_cross[0]

    # Wyyyy
    Wfit[1, 1, 1, 1] = kt_cross[1]

    # Wzzzz
    Wfit[2, 2, 2, 2] = kt_cross[2]

    # Wxxxy
    Wfit[0, 0, 0, 1] = Wfit[0, 0, 1, 0] = Wfit[0, 1, 0, 0] = kt_cross[3]
    Wfit[1, 0, 0, 0] = kt_cross[3]

    # Wxxxz
    Wfit[0, 0, 0, 2] = Wfit[0, 0, 2, 0] = Wfit[0, 2, 0, 0] = kt_cross[4]
    Wfit[2, 0, 0, 0] = kt_cross[4]

    # Wxyyy
    Wfit[0, 1, 1, 1] = Wfit[1, 0, 1, 1] = Wfit[1, 1, 1, 0] = kt_cross[5]
    Wfit[1, 1, 0, 1] = kt_cross[5]

    # Wxxxz
    Wfit[1, 1, 1, 2] = Wfit[1, 2, 1, 1] = Wfit[2, 1, 1, 1] = kt_cross[6]
    Wfit[1, 1, 2, 1] = kt_cross[6]

    # Wxzzz
    Wfit[0, 2, 2, 2] = Wfit[2, 2, 2, 0] = Wfit[2, 0, 2, 2] = kt_cross[7]
    Wfit[2, 2, 0, 2] = kt_cross[7]

    # Wyzzz
    Wfit[1, 2, 2, 2] = Wfit[2, 2, 2, 1] = Wfit[2, 1, 2, 2] = kt_cross[8]
    Wfit[2, 2, 1, 2] = kt_cross[8]

    # Wxxyy
    Wfit[0, 0, 1, 1] = Wfit[0, 1, 0, 1] = Wfit[0, 1, 1, 0] = kt_cross[9]
    Wfit[1, 0, 0, 1] = Wfit[1, 0, 1, 0] = Wfit[1, 1, 0, 0] = kt_cross[9]

    # Wxxzz
    Wfit[0, 0, 2, 2] = Wfit[0, 2, 0, 2] = Wfit[0, 2, 2, 0] = kt_cross[10]
    Wfit[2, 0, 0, 2] = Wfit[2, 0, 2, 0] = Wfit[2, 2, 0, 0] = kt_cross[10]

    # Wyyzz
    Wfit[1, 1, 2, 2] = Wfit[1, 2, 1, 2] = Wfit[1, 2, 2, 1] = kt_cross[11]
    Wfit[2, 1, 1, 2] = Wfit[2, 2, 1, 1] = Wfit[2, 1, 2, 1] = kt_cross[11]

    # Wxxyz
    Wfit[0, 0, 1, 2] = Wfit[0, 0, 2, 1] = Wfit[0, 1, 0, 2] = kt_cross[12]
    Wfit[0, 1, 2, 0] = Wfit[0, 2, 0, 1] = Wfit[0, 2, 1, 0] = kt_cross[12]
    Wfit[1, 0, 0, 2] = Wfit[1, 0, 2, 0] = Wfit[1, 2, 0, 0] = kt_cross[12]
    Wfit[2, 0, 0, 1] = Wfit[2, 0, 1, 0] = Wfit[2, 1, 0, 0] = kt_cross[12]

    # Wxyyz
    Wfit[0, 1, 1, 2] = Wfit[0, 1, 2, 1] = Wfit[0, 2, 1, 1] = kt_cross[13]
    Wfit[1, 0, 1, 2] = Wfit[1, 1, 0, 2] = Wfit[1, 1, 2, 0] = kt_cross[13]
    Wfit[1, 2, 0, 1] = Wfit[1, 2, 1, 0] = Wfit[2, 0, 1, 1] = kt_cross[13]
    Wfit[2, 1, 0, 1] = Wfit[2, 1, 1, 0] = Wfit[1, 0, 2, 1] = kt_cross[13]

    # Wxyzz
    Wfit[0, 1, 2, 2] = Wfit[0, 2, 1, 2] = Wfit[0, 2, 2, 1] = kt_cross[14]
    Wfit[1, 0, 2, 2] = Wfit[1, 2, 0, 2] = Wfit[1, 2, 2, 0] = kt_cross[14]
    Wfit[2, 0, 1, 2] = Wfit[2, 0, 2, 1] = Wfit[2, 1, 0, 2] = kt_cross[14]
    Wfit[2, 1, 2, 0] = Wfit[2, 2, 0, 1] = Wfit[2, 2, 1, 0] = kt_cross[14]

    # Function to be tested
    W4D = dki.Wcons(kt_cross)

    Wfit = Wfit.reshape(-1)
    W4D = W4D.reshape(-1)

    assert_array_almost_equal(W4D, Wfit)


def test_spherical_dki_statistics():
    # tests if MK, AK, RK and MSK are equal to expected values of a spherical
    # kurtosis tensor

    # Define multi voxel spherical kurtosis simulations
    MParam = np.zeros((2, 2, 2, 27))
    MParam[0, 0, 0] = MParam[0, 0, 1] = MParam[0, 1, 0] = params_sph
    MParam[0, 1, 1] = MParam[1, 1, 0] = params_sph
    # MParam[1, 1, 1], MParam[1, 0, 0], and MParam[1, 0, 1] remains zero

    MRef = np.zeros((2, 2, 2))
    MRef[0, 0, 0] = MRef[0, 0, 1] = MRef[0, 1, 0] = Kref_sphere
    MRef[0, 1, 1] = MRef[1, 1, 0] = Kref_sphere
    MRef[1, 1, 1] = MRef[1, 0, 0] = MRef[1, 0, 1] = 0

    # Mean kurtosis analytical solution
    MK_multi = mean_kurtosis(MParam, analytical=True)
    assert_array_almost_equal(MK_multi, MRef)

    # radial kurtosis analytical solution
    RK_multi = radial_kurtosis(MParam, analytical=True)
    assert_array_almost_equal(RK_multi, MRef)

    # axial kurtosis analytical solution
    AK_multi = axial_kurtosis(MParam, analytical=True)
    assert_array_almost_equal(AK_multi, MRef)

    # mean kurtosis tensor
    MSK_multi = mean_kurtosis_tensor(MParam)
    assert_array_almost_equal(MSK_multi, MRef)

    # radial kurtosis tensor
    RKT_multi = radial_tensor_kurtosis(MParam)
    assert_array_almost_equal(RKT_multi, MRef)

    # kurtosis fractional anisotropy (isotropic case kfa=0)
    KFA_multi = kurtosis_fractional_anisotropy(MParam)
    assert_array_almost_equal(KFA_multi, 0 * MRef)


def test_single_voxel_DKI_stats():
    # tests if DKI metrics are equal to expected values for a single fiber
    # simulate randomly oriented
    fie = 0.49
    ADi = 0.00099
    ADe = 0.00226
    RDi = 0
    RDe = 0.00087
    MD = fie * (ADi + 2 * RDi) / 3 + (1 - fie) * (ADe + 2 * RDe) / 3
    # Reference values
    AD = fie * ADi + (1 - fie) * ADe
    AK = 3 * fie * (1 - fie) * ((ADi - ADe) / AD) ** 2
    RD = fie * RDi + (1 - fie) * RDe
    RK = 3 * fie * (1 - fie) * ((RDi - RDe) / RD) ** 2
    MKT = (
        3
        * fie
        * (1 - fie)
        * (RDe**2 + (ADe - ADi - RDe) * (7 * RDe + 3 * (ADe - ADi)) / 15)
        / (MD**2)
    )
    ref_vals = np.array([AD, AK, RD, RK, RK, MKT])

    # simulate fiber randomly oriented
    theta = random.uniform(0, 180)
    phi = random.uniform(0, 320)
    angles = [(theta, phi), (theta, phi)]
    mevals = np.array([[ADi, RDi, RDi], [ADe, RDe, RDe]])
    frac = [fie * 100, (1 - fie) * 100]
    signal, dt, kt = multi_tensor_dki(
        gtab_2s, mevals, S0=100, angles=angles, fractions=frac, snr=None
    )
    evals, evecs = decompose_tensor(from_lower_triangular(dt))
    dki_par = np.concatenate((evals, evecs[0], evecs[1], evecs[2], kt), axis=0)

    # Estimates using dki functions
    ADe1 = dti.axial_diffusivity(evals)
    RDe1 = dti.radial_diffusivity(evals)
    AKe1 = axial_kurtosis(dki_par)
    RKe1 = radial_kurtosis(dki_par)
    RTKe1 = radial_tensor_kurtosis(dki_par)
    MKTe1 = mean_kurtosis_tensor(dki_par)
    e1_vals = np.array([ADe1, AKe1, RDe1, RKe1, RTKe1, MKTe1])
    assert_array_almost_equal(e1_vals, ref_vals)

    # Estimates using the kurtosis class object
    dkiM = dki.DiffusionKurtosisModel(gtab_2s)
    dkiF = dkiM.fit(signal)
    e2_vals = np.array([dkiF.ad, dkiF.ak(), dkiF.rd, dkiF.rk(), dkiF.rtk(), dkiF.mkt()])
    assert_array_almost_equal(e2_vals, ref_vals)

    # test MK (note this test correspond to the MK singularity L2==L3)
    MK_as = dkiF.mk()
    sph = Sphere(xyz=gtab.bvecs[gtab.bvals > 0])
    MK_nm = np.mean(dkiF.akc(sph))

    assert_array_almost_equal(MK_as, MK_nm, decimal=1)


def test_compare_analytical_and_numerical_methods():
    # tests if analytical solution of MK/RK/AK produces the same results than
    # their respective numerical methods

    # DKI Model fitting
    dkiM = dki.DiffusionKurtosisModel(gtab_2s)
    dkiF = dkiM.fit(signal_cross)

    # MK analytical and numerical solution
    MK_as = dkiF.rk(analytical=True)
    MK_nm = dkiF.rk(analytical=False)
    assert_array_almost_equal(MK_as, MK_nm)

    # RK analytical and numerical solution
    RK_as = dkiF.rk(analytical=True)
    RK_nm = dkiF.rk(analytical=False)
    assert_array_almost_equal(RK_as, RK_nm)

    # AK analytical and numerical solution
    AK_as = dkiF.ak(analytical=True)
    AK_nm = dkiF.ak(analytical=False)
    assert_array_almost_equal(AK_as, AK_nm)


def test_MK_singularities():
    # To test MK in case that analytical solution was a singularity not covered
    # by other tests

    dkiM = dki.DiffusionKurtosisModel(gtab_2s)

    # test singularity L1 == L2 - this is the case of a prolate diffusion
    # tensor for crossing fibers at 90 degrees
    angles_all = np.array(
        [[(90, 0), (90, 0), (0, 0), (0, 0)], [(89.9, 0), (89.9, 0), (0, 0), (0, 0)]]
    )
    for angles_90 in angles_all:
        s_90, dt_90, kt_90 = multi_tensor_dki(
            gtab_2s,
            mevals_cross,
            S0=100,
            angles=angles_90,
            fractions=frac_cross,
            snr=None,
        )
        dkiF = dkiM.fit(s_90)
        MK_an = dkiF.mk(analytical=True)
        MK_nm = dkiF.mk(analytical=False)

        assert_almost_equal(MK_an, MK_nm, decimal=3)

        # test singularity L1 == L3 and L1 != L2
        # since L1 is defined as the larger eigenvalue and L3 the smallest
        # eigenvalue, this singularity theoretically will never be called,
        # because for L1 == L3, L2 have also to be  = L1 and L2.
        # Nevertheless, I decided to include this test since this singularity
        # is relevant for cases that eigenvalues are not ordered

        # artificially revert the eigenvalue and eigenvector order
        dki_params = dkiF.model_params.copy()
        dki_params[1] = dkiF.model_params[2]
        dki_params[2] = dkiF.model_params[1]
        dki_params[4] = dkiF.model_params[5]
        dki_params[5] = dkiF.model_params[4]
        dki_params[7] = dkiF.model_params[8]
        dki_params[8] = dkiF.model_params[7]
        dki_params[10] = dkiF.model_params[11]
        dki_params[11] = dkiF.model_params[10]

        MK_an = dki.mean_kurtosis(dki_params, analytical=True)
        MK_nm = dki.mean_kurtosis(dki_params, analytical=False)

        assert_almost_equal(MK_an, MK_nm, decimal=3)


def test_dki_errors():
    # first error of DKI module is if a unknown fit method is given
    assert_raises(ValueError, dki.DiffusionKurtosisModel, gtab_2s, fit_method="JOANA")

    # second error of DKI module is if a min_signal is defined as negative
    assert_raises(ValueError, dki.DiffusionKurtosisModel, gtab_2s, min_signal=-1)
    # try case with correct min_signal
    dkiM = dki.DiffusionKurtosisModel(gtab_2s, min_signal=1)
    dkiF = dkiM.fit(DWI)
    assert_array_almost_equal(dkiF.model_params, multi_params)

    # third error is if a given mask do not have same shape as data
    dkiM = dki.DiffusionKurtosisModel(gtab_2s)

    # test a correct mask
    dkiF = dkiM.fit(DWI)
    mask_correct = dkiF.fa > 0
    mask_correct[1, 1] = False
    multi_params[1, 1] = np.zeros(27)
    dkiF = dkiM.fit(DWI, mask=mask_correct)
    assert_array_almost_equal(dkiF.model_params, multi_params)
    # test incorrect mask ("multi-voxel" fit types)
    mask_not_correct = np.array([[True, True, False], [True, False, False]])
    assert_raises(ValueError, dkiM.fit, DWI, mask=mask_not_correct)
    # test incorrect mask ("single-voxel" fit types)
    if have_cvxpy:
        dkiM = dki.DiffusionKurtosisModel(gtab_2s, fit_method="NLS")
        assert_raises(ValueError, dkiM.fit, DWI, mask=mask_not_correct)

        # error if data with only one non zero b-value is given
        assert_raises(ValueError, dki.DiffusionKurtosisModel, gtab)
    else:
        assert_raises(
            TripWireError, dki.DiffusionKurtosisModel, gtab_2s, fit_method="NLS"
        )

    # Extra checks for CLS fitting
    if have_cvxpy:
        assert_raises(
            ValueError,
            dki.DiffusionKurtosisModel,
            gtab_2s,
            fit_method="CLS",
            convexity_level="all",
        )
        assert_raises(
            ValueError,
            dki.DiffusionKurtosisModel,
            gtab_2s,
            fit_method="CLS",
            convexity_level=3,
        )
        # Check that maximum convexity levels is set to 4,
        # when large one is given

        with warnings.catch_warnings(record=True) as l_warns:
            dkim = dki.DiffusionKurtosisModel(
                gtab_2s, fit_method="CLS", convexity_level=6
            )
            check_for_warnings(l_warns, "Maximum convexity_level supported is 4.")
            assert_almost_equal(dkim.convexity_level, 4)


def test_kurtosis_maximum():
    # TEST 1
    # simulate a crossing fibers intersecting at 70 degrees. The first fiber
    # is aligned to the x-axis while the second fiber is aligned to the x-z
    # plane with an angular deviation of 70 degrees from the first one.
    # According to Neto Henriques et al, 2015 (NeuroImage 111: 85-99), the
    # kurtosis tensor of this simulation will have a maxima aligned to axis y
    angles = [(90, 0), (90, 0), (20, 0), (20, 0)]
    signal_70, dt_70, kt_70 = multi_tensor_dki(
        gtab_2s, mevals_cross, S0=100, angles=angles, fractions=frac_cross, snr=None
    )
    # prepare inputs
    dkiM = dki.DiffusionKurtosisModel(gtab_2s, fit_method="WLS")
    dkiF = dkiM.fit(signal_70)
    MD = dkiF.md
    kt = dkiF.kt
    R = dkiF.evecs
    evals = dkiF.evals
    dt = lower_triangular(np.dot(np.dot(R, np.diag(evals)), R.T))
    sphere = default_sphere

    # compute maxima
    k_max_cross, max_dir = dki._voxel_kurtosis_maximum(dt, MD, kt, sphere, gtol=1e-5)

    yaxis = np.array([0.0, 1.0, 0.0])
    cos_angle = np.abs(np.dot(max_dir[0], yaxis))
    assert_almost_equal(cos_angle, 1.0)

    # TEST 2
    # test the function on cases of well aligned fibers oriented in a random
    # defined direction. According to Neto Henriques et al, 2015 (NeuroImage
    # 111: 85-99), the maxima of kurtosis is any direction perpendicular to the
    # fiber direction. Moreover, according to multicompartmetal simulations,
    # kurtosis in this direction has to be equal to:
    fie = 0.49
    ADi = 0.00099
    ADe = 0.00226
    RDi = 0
    RDe = 0.00087
    RD = fie * RDi + (1 - fie) * RDe
    RK = 3 * fie * (1 - fie) * ((RDi - RDe) / RD) ** 2

    # prepare simulation:
    theta = random.uniform(0, 180)
    phi = random.uniform(0, 320)
    angles = [(theta, phi), (theta, phi)]
    mevals = np.array([[ADi, RDi, RDi], [ADe, RDe, RDe]])
    frac = [fie * 100, (1 - fie) * 100]
    signal, dt, kt = multi_tensor_dki(
        gtab_2s, mevals, angles=angles, fractions=frac, snr=None
    )

    # prepare inputs
    dkiM = dki.DiffusionKurtosisModel(gtab_2s, fit_method="WLS")
    dkiF = dkiM.fit(signal)
    MD = dkiF.md
    kt = dkiF.kt
    R = dkiF.evecs
    evals = dkiF.evals
    dt = lower_triangular(np.dot(np.dot(R, np.diag(evals)), R.T))

    # compute maxima
    k_max, max_dir = dki._voxel_kurtosis_maximum(dt, MD, kt, sphere, gtol=1e-5)

    # check if max direction is perpendicular to fiber direction
    fdir = np.array([sphere2cart(1.0, np.deg2rad(theta), np.deg2rad(phi))])
    cos_angle = np.abs(np.dot(max_dir[0], fdir[0]))
    assert_almost_equal(cos_angle, 0.0, decimal=5)

    # check if max direction is equal to expected value
    assert_almost_equal(k_max, RK)
    assert_almost_equal(dkiF.kmax(sphere=sphere, gtol=1e-5), RK)

    # According to Neto Henriques et al., 2015 (NeuroImage 111: 85-99),
    # e.g. see figure 1 of this article, kurtosis maxima for the first test is
    # also equal to the maxima kurtosis value of well-aligned fibers, since
    # simulations parameters (apart from fiber directions) are equal
    assert_almost_equal(k_max_cross, RK)

    # Test 3 - Test performance when kurtosis is spherical - this case, can be
    # problematic since a spherical kurtosis does not have an maximum
    k_max, max_dir = dki._voxel_kurtosis_maximum(
        dt_sph, np.mean(evals_sph), kt_sph, sphere, gtol=1e-2
    )
    assert_almost_equal(k_max, Kref_sphere)

    # Test 4 - Test performance when kt have all elements zero - this case, can
    # be problematic this case does not have an maximum
    k_max, max_dir = dki._voxel_kurtosis_maximum(
        dt_sph, np.mean(evals_sph), np.zeros(15), sphere, gtol=1e-2
    )
    assert_almost_equal(k_max, 0.0)


def test_multi_voxel_kurtosis_maximum():
    # Multi-voxel simulations parameters
    FIE = np.array([[[0.30, 0.32], [0.74, 0.51]], [[0.47, 0.21], [0.80, 0.63]]])
    RDI = np.zeros((2, 2, 2))
    ADI = np.array(
        [[[1e-3, 1.3e-3], [0.8e-3, 1e-3]], [[0.9e-3, 0.99e-3], [0.89e-3, 1.1e-3]]]
    )
    ADE = np.array(
        [[[2.2e-3, 2.3e-3], [2.8e-3, 2.1e-3]], [[1.9e-3, 2.5e-3], [1.89e-3, 2.1e-3]]]
    )
    Tor = np.array([[[2.6, 2.4], [2.8, 2.1]], [[2.9, 2.5], [2.7, 2.3]]])
    RDE = ADE / Tor

    # prepare simulation:
    DWIsim = np.zeros((2, 2, 2, gtab_2s.bvals.size))

    for i in range(2):
        for j in range(2):
            for k in range(2):
                ADi = ADI[i, j, k]
                RDi = RDI[i, j, k]
                ADe = ADE[i, j, k]
                RDe = RDE[i, j, k]
                fie = FIE[i, j, k]
                mevals = np.array([[ADi, RDi, RDi], [ADe, RDe, RDe]])
                frac = [fie * 100, (1 - fie) * 100]
                theta = random.uniform(0, 180)
                phi = random.uniform(0, 320)
                angles = [(theta, phi), (theta, phi)]
                signal, dt, kt = multi_tensor_dki(
                    gtab_2s, mevals, angles=angles, fractions=frac, snr=None
                )
                DWIsim[i, j, k, :] = signal

    # Ground truth Maximum kurtosis
    RD = FIE * RDI + (1 - FIE) * RDE
    RK = 3 * FIE * (1 - FIE) * ((RDI - RDE) / RD) ** 2

    # prepare inputs
    dkiM = dki.DiffusionKurtosisModel(gtab_2s, fit_method="WLS")
    dkiF = dkiM.fit(DWIsim)

    # TEST - when no sphere is given
    k_max = dki.kurtosis_maximum(dkiF.model_params)
    assert_almost_equal(k_max, RK, decimal=4)

    # TEST - when sphere is given
    k_max = dki.kurtosis_maximum(dkiF.model_params, sphere=default_sphere)
    assert_almost_equal(k_max, RK, decimal=4)

    # TEST - when mask is given
    mask = np.ones((2, 2, 2), dtype="bool")
    mask[1, 1, 1] = 0
    RK[1, 1, 1] = 0
    k_max = dki.kurtosis_maximum(dkiF.model_params, mask=mask)
    assert_almost_equal(k_max, RK, decimal=4)

    # TEST if wrong mask is given
    mask_not_correct = np.array([[True, True, False], [True, False, False]])
    assert_raises(
        ValueError, dki.kurtosis_maximum, dkiF.model_params, mask=mask_not_correct
    )


def test_kurtosis_fa():
    # KFA = sqrt(4/5) if kurtosis is non-zero only in one direction
    mevals = np.array([[0.002, 0, 0], [0.003, 0, 0]])
    angles = [(45, 0), (45, 0)]
    fie = 0.5
    frac = [fie * 100, (1 - fie) * 100]
    signal, dt, kt = multi_tensor_dki(
        gtab_2s, mevals, S0=100, angles=angles, fractions=frac, snr=None
    )

    dkiM = dki.DiffusionKurtosisModel(gtab_2s)
    dkiF = dkiM.fit(signal)
    assert_almost_equal(dkiF.kfa, np.sqrt(4 / 5))

    # KFA = sqrt(13/5) for systems of two tensors with same AD and RD values
    # See appendix of Gleen et al., 2015 Quantitative assessment of diffusional
    # kurtosis anisotropy. NMR Biomed 28; 448-459. doi:10.1002/nbm.3271
    mevals = np.array([[0.003, 0.001, 0.001], [0.003, 0.001, 0.001]])
    angles = [(40, -10), (-45, 10)]
    fie = 0.5
    frac = [fie * 100, (1 - fie) * 100]
    signal, dt, kt = multi_tensor_dki(
        gtab_2s, mevals, S0=100, angles=angles, fractions=frac, snr=None
    )

    dkiM = dki.DiffusionKurtosisModel(gtab_2s)
    dkiF = dkiM.fit(signal)
    assert_almost_equal(dkiF.kfa, np.sqrt(13 / 15))

    # KFA = 0 if MKT = 0
    mevals = np.array([[0.003, 0.001, 0.001], [0.003, 0.001, 0.001]])
    angles = [(40, -10), (40, -10)]
    fie = 0.5
    frac = [fie * 100, (1 - fie) * 100]
    signal, dt, kt = multi_tensor_dki(
        gtab_2s, mevals, S0=100, angles=angles, fractions=frac, snr=None
    )

    dkiM = dki.DiffusionKurtosisModel(gtab_2s)
    dkiF = dkiM.fit(signal)
    assert_almost_equal(dkiF.kfa, 0)



# cython: boundscheck=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: wraparound=False
# cython: nonecheck=False
# cython: overflowcheck=False


from dipy.align.fused_types cimport floating


cdef floating f_array_min(floating* arr, int n) noexcept nogil:
    """Return the minimum value of an array."""
    cdef int i
    cdef double min_val = arr[0]
    for i in range(1, n):
        if arr[i] < min_val:
            min_val = arr[i]
    return min_val


cdef floating f_array_max(floating* arr, int n) noexcept nogil:
    """Compute the maximum value of an array."""
    cdef int i
    cdef double max_val = arr[0]
    for i in range(1, n):
        if arr[i] > max_val:
            max_val = arr[i]
    return max_val


__all__ = [
    "MinimumAverageDirectFlipMetric",
    "Metric",
    "CosineMetric",
    "AveragePointwiseEuclideanMetric",
    "EuclideanMetric",
    "dist",
    "mdf",
    "mean_manhattan_distance",
    "mean_euclidean_distance",
]

import numpy as np

from dipy.segment.metricspeed import (
    AveragePointwiseEuclideanMetric,
    CosineMetric,
    Metric,
    MinimumAverageDirectFlipMetric,
    SumPointwiseEuclideanMetric,
    dist,
)

# Creates aliases
EuclideanMetric = SumPointwiseEuclideanMetric


def mdf(s1, s2):
    """Computes the MDF (Minimum average Direct-Flip) distance between two
    streamlines.

    Streamlines must have the same number of points.

    See :footcite:p:`Garyfallidis2012a` for a definition of the distance.

    Parameters
    ----------
    s1 : 2D array
        A streamline (sequence of N-dimensional points).
    s2 : 2D array
        A streamline (sequence of N-dimensional points).

    Returns
    -------
    double
        Distance between two streamlines.

    References
    ----------
    .. footbibliography::
    """
    return dist(MinimumAverageDirectFlipMetric(), s1, s2)


def mean_manhattan_distance(a, b):
    """Compute the average Manhattan-L1 distance (MDF without flip)

    Arrays are representing a single streamline or a list of streamlines
    that have the same number of N-dimensional points (two last axis).

    Parameters
    ----------
    a : 2D or 3D array
        A streamline or concatenated streamlines
        (array of S streamlines by P points in N dimension).
    b : 2D or 3D array
        A streamline or concatenated streamlines
        (array of S streamlines by P points in N dimension).

    Returns
    -------
    1D array
        Distance between each S streamlines
    """
    return np.mean(np.sum(np.abs(a - b), axis=-1), axis=-1)


def mean_euclidean_distance(a, b):
    """Compute the average Euclidean-L2 distance (MDF without flip)

    Arrays are representing a single streamline or a list of streamlines
    that have the same number of N-dimensional points (two last axis).

    Parameters
    ----------
    a : 2D or 3D array
        A streamline or concatenated streamlines
        (array of S streamlines by P points in N dimension).
    b : 2D or 3D array
        A streamline or concatenated streamlines
        (array of S streamlines by P points in N dimension).

    Returns
    -------
    1D array
        Distance between each S streamlines
    """
    return np.mean(np.sqrt(np.sum(np.square(a - b), axis=-1)), axis=-1)



import os
import tempfile
import time
from warnings import warn

import numpy as np
from numpy.lib.stride_tricks import sliding_window_view
from tqdm import tqdm

from dipy.stats.sketching import count_sketch
from dipy.testing.decorators import warning_for_keywords
from dipy.utils.optpkg import optional_package

sklearn, has_sklearn, _ = optional_package("sklearn")
linear_model, _, _ = optional_package("sklearn.linear_model")


def _vol_split(train, vol_idx):
    """Split the 3D volumes into the train and test set.

    Parameters
    ----------
    train : numpy.ndarray
        Array of all 3D patches flattened out to be 2D.
    vol_idx : int
        The volume number that needs to be held out for training.

    Returns
    -------
    cur_x : numpy.ndarray of shape (nvolumes * patch_size) x (nvoxels)
        Array of patches corresponding to all volumes except the held out volume.
    y : numpy.ndarray of shape (patch_size) x (nvoxels)
        Array of patches corresponding to the volume that is used a target for
        denoising.

    """
    mask = np.zeros(train.shape[0], dtype=bool)
    mask[vol_idx] = True
    cur_x = train[~mask].reshape((train.shape[0] - 1) * train.shape[1], train.shape[2])
    y = train[vol_idx, train.shape[1] // 2, :]
    return cur_x, y


def _extract_3d_patches(arr, patch_radius):
    """Extract 3D patches from 4D DWI data.

    Parameters
    ----------
    arr : ndarray
        The 4D noisy DWI data to be denoised.
    patch_radius : int or array of shape (3,)
        The radius of the local patch to be taken around each voxel (in
        voxels).

    Returns
    -------
    all_patches : ndarray
        All 3D patches flattened out to be 2D corresponding to the each 3D
        volume of the 4D DWI data.

    """
    patch_radius = np.asarray(patch_radius, dtype=int)
    if patch_radius.size == 1:
        patch_radius = np.repeat(patch_radius, 3)
    elif patch_radius.size != 3:
        raise ValueError("patch_radius should have length 1 or 3")

    patch_size = 2 * patch_radius + 1
    dim = arr.shape[-1]

    # Calculate the shape of the output array
    output_shape = tuple(arr.shape[i] - 2 * patch_radius[i] for i in range(3))
    total_patches = np.prod(output_shape)

    patches = sliding_window_view(arr, tuple(patch_size) + (dim,))

    # Reshape and transpose the patches to match the original function's output shape
    all_patches = patches.reshape(total_patches, np.prod(patch_size), dim)
    all_patches = all_patches.transpose(2, 1, 0)

    return np.array(all_patches)


def _fit_denoising_model(train, vol_idx, model, alpha):
    """Fit a single 3D volume using a train and test phase.

    Parameters
    ----------
    train : ndarray
        Array of all 3D patches flattened out to be 2D.
    vol_idx : int
        The volume number that needs to be held out for training.
    model : str or sklearn.base.RegressorMixin
        This will determine the algorithm used to solve the set of linear
        equations underlying this model. If it is a string it needs to be
        one of the following: {'ols', 'ridge', 'lasso'}. Otherwise,
        it can be an object that inherits from
        `dipy.optimize.SKLearnLinearSolver` or an object with a similar
        interface from Scikit-Learn:
        `sklearn.linear_model.LinearRegression`,
        `sklearn.linear_model.Lasso` or `sklearn.linear_model.Ridge`
        and other objects that inherit from `sklearn.base.RegressorMixin`.
    alpha : float
        Regularization parameter only for ridge and lasso regression models.
    version : int
        Version 1 or 3 of Patch2Self to use.

    Returns
    -------
    model_instance : fitted linear model object
        The fitted model instance if version is 3.
    cur_x : ndarray
        The patches corresponding to all volumes except the held out volume.

    """
    if isinstance(model, str):
        if model.lower() == "ols":
            model_instance = linear_model.Ridge(copy_X=False, alpha=1e-10)
        elif model.lower() == "ridge":
            model_instance = linear_model.Ridge(copy_X=False, alpha=alpha)
        elif model.lower() == "lasso":
            model_instance = linear_model.Lasso(copy_X=False, max_iter=50, alpha=alpha)
        else:
            raise ValueError(
                f"Invalid model string: {model}. Should be 'ols', 'ridge', or 'lasso'."
            )
    elif isinstance(model, linear_model.BaseEstimator):
        model_instance = model
    else:
        raise ValueError(
            "Model should either be a string or \
                an instance of sklearn.linear_model BaseEstimator."
        )
    cur_x, y = _vol_split(train, vol_idx)
    model_instance.fit(cur_x.T, y.T)
    return model_instance, cur_x


def vol_denoise(
    data_dict, b0_idx, dwi_idx, model, alpha, b0_denoising, verbose, tmp_dir
):
    """Denoise a single 3D volume using train and test phase.

    Parameters
    ----------
    data_dict : dict
        Dictionary containing the following:
            data_name : str
                The name of the memmap file containing the memmaped data.
            data_dtype : dtype
                The dtype of the data.
            data_shape : tuple
                The shape of the data.
            data_b0s : ndarray
                Array of all 3D patches flattened out to be 2D for b0 volumes.
            data_dwi : ndarray
                Array of all 3D patches flattened out to be 2D for dwi volumes.
    b0_idx : ndarray
        The indices of the b0 volumes.
    dwi_idx : ndarray
        The indices of the dwi volumes.
    model : sklearn.base.RegressorMixin
        This is the model that is initialized from the `_fit_denoising_model` function.
    alpha : float
        Regularization parameter only for ridge and lasso regression models.
    b0_denoising : bool
        Skips denoising b0 volumes if set to False.
    verbose : bool
        Show progress of Patch2Self and time taken.
    tmp_dir : str
        The directory to save the temporary files.

    Returns
    -------
    denoised_arr.name : str
        The name of the memmap file containing the denoised array.
    denoised_arr.dtype : dtype
        The dtype of the denoised array.
    denoised_arr.shape : tuple
        The shape of the denoised array.

    """
    data_shape = data_dict["data"][2]
    data_tmp = np.memmap(
        data_dict["data"][0],
        dtype=data_dict["data"][1],
        mode="r",
        shape=data_dict["data"][2],
    ).reshape(np.prod(data_shape[:-1]), data_shape[-1])
    data_b0s = data_dict["data_b0s"]
    data_dwi = data_dict["data_dwi"]
    p = data_tmp.shape[0] // 10
    b0_counter = 0
    dwi_counter = 0
    start_idx = 0
    denoised_arr_file = tempfile.NamedTemporaryFile(
        delete=False, dir=tmp_dir, suffix="denoised_arr"
    )
    denoised_arr_file.close()
    denoised_arr = np.memmap(
        denoised_arr_file.name, dtype=data_tmp.dtype, mode="w+", shape=data_shape
    )
    idx_counter = 0
    full_result = np.empty(
        (data_shape[0], data_shape[1], data_shape[2], data_shape[3] // 5)
    )
    b0_idx = b0_idx
    dwi_idx = dwi_idx
    if data_b0s.shape[0] == 1:
        b0_denoising = False
    if not b0_denoising:
        if verbose:
            print("b0 denoising skipped....")
    for vol_idx in tqdm(
        range(data_shape[-1]), desc="Fitting and Denoising", leave=False
    ):
        if vol_idx in b0_idx.flatten():
            if b0_denoising:
                b_fit, _ = _fit_denoising_model(data_b0s, b0_counter, model, alpha)
                b_matrix = np.zeros(data_tmp.shape[-1])
                b_fit_coef = np.insert(b_fit.coef_, b0_counter, 0)
                np.put(b_matrix, b0_idx, b_fit_coef)
                result = np.zeros(data_tmp.shape[0])
                for z in range(0, data_tmp.shape[0], p):
                    end_idx = z + p
                    if end_idx > z + p:
                        end_idx = data_tmp.shape[0]
                    result[z:end_idx] = (
                        np.matmul(np.squeeze(data_tmp[z:end_idx, :]), b_matrix)
                        + b_fit.intercept_
                    )
                full_result[..., idx_counter] = result.reshape(
                    data_shape[0], data_shape[1], data_shape[2]
                )
                idx_counter += 1
                b0_counter += 1
                del b_fit_coef
                del b_matrix
                del result
            else:
                full_result[..., idx_counter] = data_tmp[..., b0_counter].reshape(
                    data_shape[0], data_shape[1], data_shape[2]
                )
                b0_counter += 1
                idx_counter += 1
        else:
            dwi_fit, _ = _fit_denoising_model(data_dwi, dwi_counter, model, alpha)
            b_matrix = np.zeros(data_tmp.shape[-1])
            dwi_fit_coef = np.insert(dwi_fit.coef_, dwi_counter, 0)
            np.put(b_matrix, dwi_idx, dwi_fit_coef)
            del dwi_fit_coef
            result = np.zeros(data_tmp.shape[0])
            for z in range(0, data_tmp.shape[0], p):
                end_idx = z + p
                if end_idx > z + p:
                    end_idx = data_tmp.shape[0]
                result[z:end_idx] = (
                    np.matmul(np.squeeze(data_tmp[z:end_idx, :]), b_matrix)
                    + dwi_fit.intercept_
                )
            full_result[..., idx_counter] = result.reshape(
                data_shape[0], data_shape[1], data_shape[2]
            )
            idx_counter += 1
            dwi_counter += 1
        if idx_counter >= data_shape[-1] // 5:
            denoised_arr[..., start_idx : vol_idx + 1] = full_result
            start_idx = vol_idx + 1
            idx_counter = 0
    denoised_arr_idx = data_shape[-1] - data_shape[-1] % 5
    full_result_idx = full_result.shape[-1] - data_shape[-1] % 5
    denoised_arr[..., denoised_arr_idx:] = full_result[..., full_result_idx:]
    del full_result
    return denoised_arr_file.name, denoised_arr.dtype, denoised_arr.shape


@warning_for_keywords()
def patch2self(
    data,
    bvals,
    *,
    patch_radius=(0, 0, 0),
    model="ols",
    b0_threshold=50,
    out_dtype=None,
    alpha=1.0,
    verbose=False,
    b0_denoising=True,
    clip_negative_vals=False,
    shift_intensity=True,
    tmp_dir=None,
    version=3,
):
    """Patch2Self Denoiser.

    See :footcite:p:`Fadnavis2020` for further details about the method.
    See :footcite:p:`Fadnavis2024` for further details about the new method.

    Parameters
    ----------
    data : ndarray
        The 4D noisy DWI data to be denoised.
    bvals : array of shape (N,)
        Array of the bvals from the DWI acquisition
    patch_radius : int or array of shape (3,), optional
        The radius of the local patch to be taken around each voxel (in
        voxels). Default: 0 (denoise in blocks of 1x1x1 voxels).
    model : string, or sklearn.base.RegressorMixin
        This will determine the algorithm used to solve the set of linear
        equations underlying this model. If it is a string it needs to be
        one of the following: {'ols', 'ridge', 'lasso'}. Otherwise,
        it can be an object that inherits from
        `dipy.optimize.SKLearnLinearSolver` or an object with a similar
        interface from Scikit-Learn:
        `sklearn.linear_model.LinearRegression`,
        `sklearn.linear_model.Lasso` or `sklearn.linear_model.Ridge`
        and other objects that inherit from `sklearn.base.RegressorMixin`.
    b0_threshold : int, optional
        Threshold for considering volumes as b0.
    out_dtype : str or dtype, optional
        The dtype for the output array. Default: output has the same dtype as
        the input.
    alpha : float, optional
        Regularization parameter only for ridge regression model.
    verbose : bool, optional
        Show progress of Patch2Self and time taken.
    b0_denoising : bool, optional
        Skips denoising b0 volumes if set to False.
    clip_negative_vals : bool, optional
        Sets negative values after denoising to 0 using `np.clip`.
    shift_intensity : bool, optional
        Shifts the distribution of intensities per volume to give
        non-negative values.
    tmp_dir : str, optional
        The directory to save the temporary files. If None, the temporary
        files are saved in the system's default temporary directory. Default: None.
    version : int, optional
        Version 1 or 3 of Patch2Self to use. Default: 3

    Returns
    -------
    denoised array : ndarray
        This is the denoised array of the same size as that of the input data,
        clipped to non-negative values.

    References
    ----------
    .. footbibliography::

    """

    out_dtype, tmp_dir, patch_radius = _validate_inputs(
        data, out_dtype, patch_radius, version, tmp_dir
    )

    if version == 1:
        return _patch2self_version1(
            data,
            bvals,
            patch_radius,
            model,
            b0_threshold,
            out_dtype,
            alpha,
            verbose,
            b0_denoising,
            clip_negative_vals,
            shift_intensity,
        )
    return _patch2self_version3(
        data,
        bvals,
        model,
        b0_threshold,
        out_dtype,
        alpha,
        verbose,
        b0_denoising,
        clip_negative_vals,
        shift_intensity,
        tmp_dir,
    )


def _validate_inputs(data, out_dtype, patch_radius, version, tmp_dir):
    """Validate inputs for patch2self function.

    Parameters
    ----------
    data : ndarray
        The 4D noisy DWI data to be denoised.
    out_dtype : str or dtype
        The dtype for the output array.
    patch_radius : int or array of shape (3,)
        The radius of the local patch to be taken around each voxel (in
        voxels).
    version : int
        Version 1 or 3 of Patch2Self to use.
    tmp_dir : str
        The directory to save the temporary files. If None, the temporary
        files are saved in the system's default temporary directory.

    Raises
    ------
    ValueError
        If temporary directory is not None for Patch2Self version 1.
        If the patch_radius is not 0 for Patch2Self version 3.
        If the temporary directory does not exist.
        If the input data is not a 4D array.

    Warns
    -----
    If the input data has less than 10 3D volumes.

    Returns
    -------
    out_dtype : str or dtype
        The dtype for the output array.
    tmp_dir : str
        The directory to save the temporary files. If None, the temporary
        files are saved in the system's default temporary directory.

    """
    if out_dtype is None:
        out_dtype = data.dtype

    if tmp_dir is None and version == 3:
        tmp_dir = tempfile.gettempdir()

    if version not in [1, 3]:
        raise ValueError("Invalid version. Should be 1 or 3.")

    if version == 1 and tmp_dir is not None:
        raise ValueError(
            "Temporary directory is not supported for Patch2Self version 1. \
                Please set tmp_dir to None."
        )
    if patch_radius != (0, 0, 0) and version == 3:
        raise ValueError(
            "Patch radius is not supported for Patch2Self version 3. \
                Please do not set patch_radius."
        )

    if isinstance(patch_radius, list) and len(patch_radius) == 1:
        patch_radius = (patch_radius[0], patch_radius[0], patch_radius[0])

    if isinstance(patch_radius, int):
        patch_radius = (patch_radius, patch_radius, patch_radius)

    if version == 3 and tmp_dir is not None and not os.path.exists(tmp_dir):
        raise ValueError("The temporary directory does not exist.")
    if data.ndim != 4:
        raise ValueError("Patch2Self can only denoise on 4D arrays.", data.shape)
    if data.shape[3] < 10:
        warn(
            "The input data has less than 10 3D volumes. \
                Patch2Self may not give optimal denoising performance.",
            stacklevel=2,
        )

    return out_dtype, tmp_dir, patch_radius


def _patch2self_version1(
    data,
    bvals,
    patch_radius,
    model,
    b0_threshold,
    out_dtype,
    alpha,
    verbose,
    b0_denoising,
    clip_negative_vals,
    shift_intensity,
):
    """Patch2Self Denoiser.

    Parameters
    ----------
    data : ndarray
        The 4D noisy DWI data to be denoised.
    bvals : array of shape (N,)
        Array of the bvals from the DWI acquisition.
    patch_radius : int or array of shape (3,)
        The radius of the local patch to be taken around each voxel (in
        voxels).
    model : string, or sklearn.base.RegressorMixin
        This will determine the algorithm used to solve the set of linear
        equations underlying this model. If it is a string it needs to be
        one of the following: {'ols', 'ridge', 'lasso'}. Otherwise,
        it can be an object that inherits from
        `dipy.optimize.SKLearnLinearSolver` or an object with a similar
        interface from Scikit-Learn:
        `sklearn.linear_model.LinearRegression`,
        `sklearn.linear_model.Lasso` or `sklearn.linear_model.Ridge`
        and other objects that inherit from `sklearn.base.RegressorMixin`.
    b0_threshold : int
        Threshold for considering volumes as b0.
    out_dtype : str or dtype
        The dtype for the output array. Default: output has the same dtype as
        the input.
    alpha : float
        Regularization parameter only for ridge regression model.
    verbose : bool
        Show progress of Patch2Self and time taken.
    b0_denoising : bool
        Skips denoising b0 volumes if set to False.
    clip_negative_vals : bool
        Sets negative values after denoising to 0 using `np.clip`.
    shift_intensity : bool
        Shifts the distribution of intensities per volume to give
        non-negative values.

    Returns
    -------
    denoised array : ndarray
        This is the denoised array of the same size as that of the input data,
        clipped to non-negative values.
    """

    # We retain float64 precision, iff the input is in this precision:
    if data.dtype == np.float64:
        calc_dtype = np.float64

    # Otherwise, we'll calculate things in float32 (saving memory)
    else:
        calc_dtype = np.float32

    original_shape = data.shape
    if 1 in data.shape and data.shape[-1] != 1:
        position = data.shape.index(1)
        data = np.concatenate((data, data, data), position)

    # Segregates volumes by b0 threshold
    b0_idx = np.argwhere(bvals <= b0_threshold)
    dwi_idx = np.argwhere(bvals > b0_threshold)

    data_b0s = np.squeeze(np.take(data, b0_idx, axis=3))
    data_dwi = np.squeeze(np.take(data, dwi_idx, axis=3))

    # create empty arrays
    denoised_b0s = np.empty(data_b0s.shape, dtype=calc_dtype)
    denoised_dwi = np.empty(data_dwi.shape, dtype=calc_dtype)

    denoised_arr = np.empty(data.shape, dtype=calc_dtype)

    if verbose is True:
        t1 = time.time()

    # if only 1 b0 volume, skip denoising it
    if data_b0s.ndim == 3 or not b0_denoising:
        if verbose:
            print("b0 denoising skipped...")
        denoised_b0s = data_b0s

    else:
        train_b0 = _extract_3d_patches(
            np.pad(
                data_b0s,
                (
                    (patch_radius[0], patch_radius[0]),
                    (patch_radius[1], patch_radius[1]),
                    (patch_radius[2], patch_radius[2]),
                    (0, 0),
                ),
                mode="constant",
            ),
            patch_radius=patch_radius,
        )

        for vol_idx in range(0, data_b0s.shape[3]):
            b0_model, cur_x = _fit_denoising_model(
                train_b0, vol_idx, model, alpha=alpha
            )

            denoised_b0s[..., vol_idx] = b0_model.predict(cur_x.T).reshape(
                data_b0s.shape[0], data_b0s.shape[1], data_b0s.shape[2]
            )

        if verbose is True:
            print("Denoised b0 Volume: ", vol_idx)
    # Separate denoising for DWI volumes
    train_dwi = _extract_3d_patches(
        np.pad(
            data_dwi,
            (
                (patch_radius[0], patch_radius[0]),
                (patch_radius[1], patch_radius[1]),
                (patch_radius[2], patch_radius[2]),
                (0, 0),
            ),
            mode="constant",
        ),
        patch_radius=patch_radius,
    )

    # Insert the separately denoised arrays into the respective empty arrays
    for vol_idx in range(0, data_dwi.shape[3]):
        dwi_model, cur_x = _fit_denoising_model(train_dwi, vol_idx, model, alpha=alpha)
        denoised_dwi[..., vol_idx] = dwi_model.predict(cur_x.T).reshape(
            data_dwi.shape[0], data_dwi.shape[1], data_dwi.shape[2]
        )

        if verbose is True:
            print("Denoised DWI Volume: ", vol_idx)

    if verbose is True:
        t2 = time.time()
        print("Total time taken for Patch2Self: ", t2 - t1, " seconds")

    if data_b0s.ndim == 3:
        denoised_arr[:, :, :, b0_idx[0][0]] = denoised_b0s
    else:
        for i, idx in enumerate(b0_idx):
            denoised_arr[:, :, :, idx[0]] = np.squeeze(denoised_b0s[..., i])

    for i, idx in enumerate(dwi_idx):
        denoised_arr[:, :, :, idx[0]] = np.squeeze(denoised_dwi[..., i])

    if 1 in original_shape and original_shape[-1] != 1:
        denoised_arr = np.take(denoised_arr, [0], axis=position)

    denoised_arr = _apply_post_processing(
        denoised_arr, shift_intensity, clip_negative_vals
    )
    return np.array(denoised_arr, dtype=out_dtype)


def _patch2self_version3(
    data,
    bvals,
    model,
    b0_threshold,
    out_dtype,
    alpha,
    verbose,
    b0_denoising,
    clip_negative_vals,
    shift_intensity,
    tmp_dir,
):
    """Patch2Self Denoiser.

    Parameters
    ----------
    data : ndarray
        The 4D noisy DWI data to be denoised.
    bvals : array of shape (N,)
        Array of the bvals from the DWI acquisition.
    model : string, or sklearn.base.RegressorMixin
        This will determine the algorithm used to solve the set of linear
        equations underlying this model. If it is a string it needs to be
        one of the following: {'ols', 'ridge', 'lasso'}. Otherwise,
        it can be an object that inherits from
        `dipy.optimize.SKLearnLinearSolver` or an object with a similar
        interface from Scikit-Learn:
        `sklearn.linear_model.LinearRegression`,
        `sklearn.linear_model.Lasso` or `sklearn.linear_model.Ridge`
        and other objects that inherit from `sklearn.base.RegressorMixin`.
    b0_threshold : int
        Threshold for considering volumes as b0.
    out_dtype : str or dtype
        The dtype for the output array. Default: output has the same dtype as
        the input.
    alpha : float
        Regularization parameter only for ridge regression model.
    verbose : bool
        Show progress of Patch2Self and time taken.
    b0_denoising : bool
        Skips denoising b0 volumes if set to False.
    clip_negative_vals : bool
        Sets negative values after denoising to 0 using `np.clip`.
    shift_intensity : bool
        Shifts the distribution of intensities per volume to give
        non-negative values.
    tmp_dir : str
        The directory to save the temporary files. If None, the temporary
        files are saved in the system's default temporary directory.

    Returns
    -------
    denoised array : ndarray
        This is the denoised array of the same size as that of the input data,
        clipped to non-negative values.

    """
    tmp_file = tempfile.NamedTemporaryFile(delete=False, dir=tmp_dir, suffix="tmp_file")
    tmp_file.close()
    tmp = np.memmap(
        tmp_file.name,
        dtype=data.dtype,
        mode="w+",
        shape=(data.shape[0], data.shape[1], data.shape[2], data.shape[3]),
    )
    p = data.shape[-1] // 5
    idx_start = 0
    for z in range(0, data.shape[3], p):
        end_idx = z + p
        if end_idx > data.shape[3]:
            end_idx = data.shape[3]
        if verbose:
            print("Loading data from {} to {}".format(idx_start, end_idx))
        tmp[..., idx_start:end_idx] = data[..., idx_start:end_idx]
        idx_start = end_idx
    sketch_rows = int(0.30 * data.shape[0] * data.shape[1] * data.shape[2])
    sketched_matrix_name, sketched_matrix_dtype, sketched_matrix_shape = count_sketch(
        tmp_file.name,
        data.dtype,
        tmp.shape,
        sketch_rows=sketch_rows,
        tmp_dir=tmp_dir,
    )
    sketched_matrix = np.memmap(
        sketched_matrix_name,
        dtype=sketched_matrix_dtype,
        mode="r",
        shape=sketched_matrix_shape,
    ).T
    if verbose:
        print("Sketching done.")
    b0_idx = np.argwhere(bvals <= b0_threshold)
    dwi_idx = np.argwhere(bvals > b0_threshold)
    data_b0s = np.take(np.squeeze(sketched_matrix), b0_idx, axis=0)
    data_dwi = np.take(np.squeeze(sketched_matrix), dwi_idx, axis=0)
    data_dict = {
        "data": [tmp_file.name, data.dtype, tmp.shape],
        "data_b0s": data_b0s,
        "data_dwi": data_dwi,
    }
    if verbose:
        t1 = time.time()
    del sketched_matrix
    os.unlink(sketched_matrix_name)
    denoised_arr_name, denoised_arr_dtype, denoised_arr_shape = vol_denoise(
        data_dict,
        b0_idx,
        dwi_idx,
        model,
        alpha,
        b0_denoising,
        verbose,
        tmp_dir=tmp_dir,
    )
    denoised_arr = np.memmap(
        denoised_arr_name,
        dtype=denoised_arr_dtype,
        mode="r+",
        shape=denoised_arr_shape,
    )
    if verbose:
        t2 = time.time()
        print("Time taken for Patch2Self: ", t2 - t1, " seconds.")

    denoised_arr = _apply_post_processing(
        denoised_arr, shift_intensity, clip_negative_vals
    )
    del tmp
    os.unlink(data_dict["data"][0])
    result = np.array(denoised_arr, dtype=out_dtype)
    del denoised_arr
    os.unlink(denoised_arr_name)
    return result


def _apply_post_processing(denoised_arr, shift_intensity, clip_negative_vals):
    """Apply post-processing steps such as clipping and shifting intensities.

    Parameters
    ----------
    denoised_arr : ndarray
        The denoised array.
    shift_intensity : bool
        Shifts the distribution of intensities per volume to give
        non-negative values.
    clip_negative_vals : bool
        Sets negative values after denoising to 0 using `np.clip`.

    Returns
    -------
    denoised_arr : ndarray
        The denoised array with post-processing applied.

    """
    if shift_intensity and not clip_negative_vals:
        for i in range(denoised_arr.shape[-1]):
            shift = np.min(denoised_arr[..., i]) - np.min(denoised_arr[..., i])
            denoised_arr[..., i] += shift
    elif clip_negative_vals and not shift_intensity:
        denoised_arr.clip(min=0, out=denoised_arr)
    elif clip_negative_vals and shift_intensity:
        warn(
            "Both `clip_negative_vals` and `shift_intensity` cannot be True. \
                Defaulting to `clip_negative_vals`...",
            stacklevel=2,
        )
        denoised_arr.clip(min=0, out=denoised_arr)
    return denoised_arr



"""Benchmarks for``dipy.reconst`` module."""

import numpy as np

from dipy.core.sphere import unique_edges
from dipy.data import default_sphere
from dipy.reconst.recspeed import local_maxima
from dipy.reconst.vec_val_sum import vec_val_vect


class BenchRecSpeed:
    def setup(self):
        vertices, faces = default_sphere.vertices, default_sphere.faces
        self.edges = unique_edges(faces)
        self.odf = np.zeros(len(vertices))
        self.odf[1] = 1.0
        self.odf[143] = 143.0
        self.odf[305] = 305.0

    def time_local_maxima(self):
        local_maxima(self.odf, self.edges)


class BenchVecValSum:
    def setup(self):
        def make_vecs_vals(shape):
            return (
                np.random.randn(*shape),
                np.random.randn(*(shape[:-2] + shape[-1:])),
            )

        shape = (10, 12, 3, 3)
        self.evecs, self.evals = make_vecs_vals(shape)

    def time_vec_val_vect(self):
        vec_val_vect(self.evecs, self.evals)


# class BenchCSD:

#     def setup(self):
#         img, self.gtab, labels_img = read_stanford_labels()
#         data = img.get_fdata()

#         labels = labels_img.get_fdata()
#         shape = labels.shape
#         mask = np.isin(labels, [1, 2])
#         mask.shape = shape

#         center = (50, 40, 40)
#         width = 12
#         a, b, c = center
#         hw = width // 2
#         idx = (slice(a - hw, a + hw), slice(b - hw, b + hw),
#                slice(c - hw, c + hw))

#         self.data_small = data[idx].copy()
#         self.mask_small = mask[idx].copy()
#         voxels = self.mask_small.sum()
#         self.small_gtab = GradientTable(self.gtab.gradients[:75])


# def time_csdeconv_basic(self):
#     # TODO: define response and remove None
#     sh_order_max = 8
#     model = ConstrainedSphericalDeconvModel(self.gtab, None,
#                                             sh_order_max=sh_order_max)
#     model.fit(self.data_small, self.mask_small)

# def time_csdeconv_small_dataset(self):
#      # TODO: define response and remove None
#     # Smaller data set
#     # data_small = data_small[..., :75].copy()
#     sh_order_max = 8
#     model = ConstrainedSphericalDeconvModel(self.small_gtab, None,
#                                             sh_order_max=sh_order_max)
#     model.fit(self.data_small, self.mask_small)

# def time_csdeconv_super_resolution(self):
#      # TODO: define response and remove None
#     # Super resolution
#     sh_order_max = 12
#     model = ConstrainedSphericalDeconvModel(self.gtab, None,
#                                             sh_order_max=sh_order_max)
#     model.fit(self.data_small, self.mask_small)



"""Testing DKI microstructure"""

import random

import numpy as np
from numpy.testing import (
    assert_,
    assert_allclose,
    assert_almost_equal,
    assert_array_almost_equal,
    assert_raises,
)
import pytest

from dipy.core.gradients import gradient_table
from dipy.data import default_sphere, get_fnames, get_sphere
from dipy.io.gradients import read_bvals_bvecs
from dipy.reconst.dki import common_fit_methods
import dipy.reconst.dki_micro as dki_micro
from dipy.reconst.dti import eig_from_lo_tri
from dipy.sims.voxel import _check_directions, multi_tensor, multi_tensor_dki
from dipy.utils.optpkg import optional_package

cvxpy, have_cvxpy, _ = optional_package("cvxpy", min_version="1.4.1")
needs_cvxpy = pytest.mark.skipif(not have_cvxpy, reason="Requires CVXPY")

gtab_2s, DWIsim, DWIsim_all_taylor = None, None, None
FIE, RDI, ADI, ADE, Tor, RDE = None, None, None, None, None, None


def setup_module():
    global gtab_2s, DWIsim, DWIsim_all_taylor, FIE, RDI, ADI, ADE, Tor, RDE

    fimg, fbvals, fbvecs = get_fnames(name="small_64D")
    bvals, bvecs = read_bvals_bvecs(fbvals, fbvecs)

    # 2 shells for techniques that requires multishell data
    bvals_2s = np.concatenate((bvals, bvals * 2), axis=0)
    bvecs_2s = np.concatenate((bvecs, bvecs), axis=0)
    gtab_2s = gradient_table(bvals_2s, bvecs=bvecs_2s)

    # single fiber simulate (which is the assumption of our model)
    FIE = np.array([[[0.30, 0.32], [0.74, 0.51]], [[0.47, 0.21], [0.80, 0.63]]])
    RDI = np.zeros((2, 2, 2))
    ADI = np.array(
        [[[1e-3, 1.3e-3], [0.8e-3, 1e-3]], [[0.9e-3, 0.99e-3], [0.89e-3, 1.1e-3]]]
    )
    ADE = np.array(
        [[[2.2e-3, 2.3e-3], [2.8e-3, 2.1e-3]], [[1.9e-3, 2.5e-3], [1.89e-3, 2.1e-3]]]
    )
    Tor = np.array([[[2.6, 2.4], [2.8, 2.1]], [[2.9, 2.5], [2.7, 2.3]]])
    RDE = ADE / Tor

    # prepare simulation:
    DWIsim = np.zeros((2, 2, 2, gtab_2s.bvals.size))

    # Diffusion microstructural model assumes that signal does not have Taylor
    # approximation components larger than the fourth order. Thus parameter
    # estimates are only equal to the ground truth values of the simulation
    # if signals taylor components larger than the fourth order are removed.
    # Signal without this taylor components can be generated using the
    # multi_tensor_dki simulations. Therefore we used this function to test the
    # expected estimates of the model.

    DWIsim_all_taylor = np.zeros((2, 2, 2, gtab_2s.bvals.size))

    # Signal with all taylor components can be simulated using the function
    # multi_tensor. Generating this signals will be useful to test the
    # prediction procedures of DKI-based microstructural model.

    for i in range(2):
        for j in range(2):
            for k in range(2):
                ADi = ADI[i, j, k]
                RDi = RDI[i, j, k]
                ADe = ADE[i, j, k]
                RDe = RDE[i, j, k]
                fie = FIE[i, j, k]
                mevals = np.array([[ADi, RDi, RDi], [ADe, RDe, RDe]])
                frac = [fie * 100, (1 - fie) * 100]
                theta = random.uniform(0, 180)
                phi = random.uniform(0, 320)
                angles = [(theta, phi), (theta, phi)]
                signal, dt, kt = multi_tensor_dki(
                    gtab_2s, mevals, angles=angles, fractions=frac, snr=None
                )
                DWIsim[i, j, k, :] = signal
                signal, sticks = multi_tensor(
                    gtab_2s, mevals, angles=angles, fractions=frac, snr=None
                )
                DWIsim_all_taylor[i, j, k, :] = signal


def teardown_module():
    global gtab_2s, DWIsim, DWIsim_all_taylor, FIE, RDI, ADI, ADE, Tor, RDE
    gtab_2s, DWIsim, DWIsim_all_taylor = None, None, None
    FIE, RDI, ADI, ADE, Tor, RDE = None, None, None, None, None, None


@needs_cvxpy
def test_fit_selection():
    for model in [
        dki_micro.KurtosisMicrostructureModel,
        dki_micro.DiffusionKurtosisModel,
    ]:
        for name, method in common_fit_methods.items():
            model_instance = model(gtab_2s, fit_method=name)

            assert model_instance.fit_method == method


def test_single_fiber_model():
    # single fiber simulate (which is the assumption of our model)
    fie = 0.49
    ADi = 0.00099
    ADe = 0.00226
    RDi = 0
    RDe = 0.00087

    # prepare simulation:
    theta = random.uniform(0, 180)
    phi = random.uniform(0, 320)
    angles = [(theta, phi), (theta, phi)]
    mevals = np.array([[ADi, RDi, RDi], [ADe, RDe, RDe]])
    frac = [fie * 100, (1 - fie) * 100]
    signal, dt, kt = multi_tensor_dki(
        gtab_2s, mevals, angles=angles, fractions=frac, snr=None
    )
    # DKI fit
    dkiM = dki_micro.DiffusionKurtosisModel(gtab_2s, fit_method="WLS")
    dkiF = dkiM.fit(signal)

    # Axonal Water Fraction
    AWF = dki_micro.axonal_water_fraction(
        dkiF.model_params, sphere=default_sphere, mask=None, gtol=1e-5
    )
    assert_almost_equal(AWF, fie)

    # Extra-cellular and intra-cellular components
    edt, idt = dki_micro.diffusion_components(dkiF.model_params, sphere=default_sphere)
    EDT = eig_from_lo_tri(edt)
    IDT = eig_from_lo_tri(idt)

    # check eigenvalues
    assert_array_almost_equal(EDT[0:3], np.array([ADe, RDe, RDe]))
    assert_array_almost_equal(IDT[0:3], np.array([ADi, RDi, RDi]))
    # first eigenvalue should be the direction of the fibers
    fiber_direction = _check_directions([(theta, phi)])
    f_norm = abs(np.dot(fiber_direction, np.array((EDT[3], EDT[6], EDT[9]))))
    assert_almost_equal(f_norm, 1.0)
    f_norm = abs(np.dot(fiber_direction, np.array((IDT[3], IDT[6], IDT[9]))))
    assert_almost_equal(f_norm, 1.0)

    # Test model and fit objects
    wmtiM = dki_micro.KurtosisMicrostructureModel(gtab_2s, fit_method="WLS")
    wmtiF = wmtiM.fit(signal)
    assert_allclose(wmtiF.awf, AWF, rtol=1e-6)
    assert_array_almost_equal(wmtiF.hindered_evals, np.array([ADe, RDe, RDe]))
    assert_array_almost_equal(wmtiF.restricted_evals, np.array([ADi, RDi, RDi]))
    assert_almost_equal(wmtiF.hindered_ad, ADe)
    assert_almost_equal(wmtiF.hindered_rd, RDe)
    assert_almost_equal(wmtiF.axonal_diffusivity, ADi)
    assert_almost_equal(wmtiF.tortuosity, ADe / RDe, decimal=4)

    # Test diffusion_components when a kurtosis tensors is associated with
    # negative kurtosis values. E.g of this cases is given below:
    dkiparams = np.array(
        [
            1.67135726e-03,
            5.03651205e-04,
            9.35365328e-05,
            -7.11167583e-01,
            6.23186820e-01,
            -3.25390313e-01,
            -1.75247376e-02,
            -4.78415563e-01,
            -8.77958674e-01,
            7.02804064e-01,
            6.18673368e-01,
            -3.51154825e-01,
            2.18384153,
            -2.76378153e-02,
            2.22893297,
            -2.68306546e-01,
            -1.28411610,
            -1.56557645e-01,
            -1.80850619e-01,
            -8.33152110e-01,
            -3.62410766e-01,
            1.57775442e-01,
            8.73775381e-01,
            2.77188975e-01,
            -3.67415502e-02,
            -1.56330984e-01,
            -1.62295407e-02,
        ]
    )
    edt, idt = dki_micro.diffusion_components(dkiparams)
    assert_(np.all(np.isfinite(edt)))


def test_wmti_model_multi_voxel():
    # DKI fit
    dkiM = dki_micro.DiffusionKurtosisModel(gtab_2s, fit_method="WLS")
    dkiF = dkiM.fit(DWIsim)

    # Axonal Water Fraction
    sphere = get_sphere()
    AWF = dki_micro.axonal_water_fraction(
        dkiF.model_params, sphere=sphere, mask=None, gtol=1e-5
    )
    assert_almost_equal(AWF, FIE)

    # Extra-cellular and intra-cellular components
    edt, idt = dki_micro.diffusion_components(dkiF.model_params, sphere=sphere)
    EDT = eig_from_lo_tri(edt)
    IDT = eig_from_lo_tri(idt)

    # check eigenvalues
    assert_array_almost_equal(EDT[..., 0], ADE, decimal=3)
    assert_array_almost_equal(EDT[..., 1], RDE, decimal=3)
    assert_array_almost_equal(EDT[..., 2], RDE, decimal=3)
    assert_array_almost_equal(IDT[..., 0], ADI, decimal=3)
    assert_array_almost_equal(IDT[..., 1], RDI, decimal=3)
    assert_array_almost_equal(IDT[..., 2], RDI, decimal=3)

    # Test methods performance when a signal with all zeros is present
    FIEc = FIE.copy()
    RDIc = RDI.copy()
    ADIc = ADI.copy()
    ADEc = ADE.copy()
    Torc = Tor.copy()
    RDEc = RDE.copy()
    DWIsimc = DWIsim.copy()

    FIEc[0, 0, 0] = 0
    RDIc[0, 0, 0] = 0
    ADIc[0, 0, 0] = 0
    ADEc[0, 0, 0] = 0
    Torc[0, 0, 0] = 0
    RDEc[0, 0, 0] = 0
    DWIsimc[0, 0, 0, :] = 0
    mask = np.ones((2, 2, 2))
    mask[0, 0, 0] = 0

    dkiF = dkiM.fit(DWIsimc)
    awf = dki_micro.axonal_water_fraction(dkiF.model_params, sphere=sphere, gtol=1e-5)
    assert_almost_equal(awf, FIEc)

    # Extra-cellular and intra-cellular components
    edt, idt = dki_micro.diffusion_components(dkiF.model_params, sphere=sphere, awf=awf)
    EDT = eig_from_lo_tri(edt)
    IDT = eig_from_lo_tri(idt)
    assert_array_almost_equal(EDT[..., 0], ADEc, decimal=3)
    assert_array_almost_equal(EDT[..., 1], RDEc, decimal=3)
    assert_array_almost_equal(EDT[..., 2], RDEc, decimal=3)
    assert_array_almost_equal(IDT[..., 0], ADIc, decimal=3)
    assert_array_almost_equal(IDT[..., 1], RDIc, decimal=3)
    assert_array_almost_equal(IDT[..., 2], RDIc, decimal=3)

    # Check when mask is given
    dkiF = dkiM.fit(DWIsim)
    awf = dki_micro.axonal_water_fraction(
        dkiF.model_params, sphere=sphere, gtol=1e-5, mask=mask
    )
    assert_almost_equal(awf, FIEc, decimal=3)

    # Extra-cellular and intra-cellular components
    edt, idt = dki_micro.diffusion_components(
        dkiF.model_params, sphere=sphere, awf=awf, mask=mask
    )
    EDT = eig_from_lo_tri(edt)
    IDT = eig_from_lo_tri(idt)
    assert_array_almost_equal(EDT[..., 0], ADEc, decimal=3)
    assert_array_almost_equal(EDT[..., 1], RDEc, decimal=3)
    assert_array_almost_equal(EDT[..., 2], RDEc, decimal=3)
    assert_array_almost_equal(IDT[..., 0], ADIc, decimal=3)
    assert_array_almost_equal(IDT[..., 1], RDIc, decimal=3)
    assert_array_almost_equal(IDT[..., 2], RDIc, decimal=3)

    # Check class object
    wmtiM = dki_micro.KurtosisMicrostructureModel(gtab_2s, fit_method="WLS")
    wmtiF = wmtiM.fit(DWIsim, mask=mask)
    assert_almost_equal(wmtiF.awf, FIEc, decimal=3)
    assert_almost_equal(wmtiF.axonal_diffusivity, ADIc, decimal=3)
    assert_almost_equal(wmtiF.hindered_ad, ADEc, decimal=3)
    assert_almost_equal(wmtiF.hindered_rd, RDEc, decimal=3)
    assert_almost_equal(wmtiF.tortuosity, Torc, decimal=3)


def test_dki_micro_predict_single_voxel():
    # single fiber simulate (which is the assumption of our model)
    fie = 0.49
    ADi = 0.00099
    ADe = 0.00226
    RDi = 0
    RDe = 0.00087

    # prepare simulation:
    theta = random.uniform(0, 180)
    phi = random.uniform(0, 320)
    angles = [(theta, phi), (theta, phi)]
    mevals = np.array([[ADi, RDi, RDi], [ADe, RDe, RDe]])
    frac = [fie * 100, (1 - fie) * 100]
    signal, dt, kt = multi_tensor_dki(
        gtab_2s, mevals, angles=angles, fractions=frac, snr=None
    )
    signal_gt, da = multi_tensor(
        gtab_2s, mevals, angles=angles, fractions=frac, snr=None
    )

    # Defined DKI microstrutural model
    dkiM = dki_micro.KurtosisMicrostructureModel(gtab_2s)

    # Fit single voxel signal
    dkiF = dkiM.fit(signal)

    # Check predict of KurtosisMicrostruturalModel
    pred = dkiM.predict(dkiF.model_params)
    assert_array_almost_equal(pred, signal_gt, decimal=4)

    pred = dkiM.predict(dkiF.model_params, S0=100)
    assert_array_almost_equal(pred, signal_gt * 100, decimal=4)

    # Check predict of KurtosisMicrostruturalFit
    pred = dkiF.predict(gtab_2s, S0=100)
    assert_array_almost_equal(pred, signal_gt * 100, decimal=4)


def test_dki_micro_predict_multi_voxel():
    dkiM = dki_micro.KurtosisMicrostructureModel(gtab_2s)
    dkiF = dkiM.fit(DWIsim)

    # Check predict of KurtosisMicrostruturalModel
    pred = dkiM.predict(dkiF.model_params)
    assert_array_almost_equal(pred, DWIsim_all_taylor, decimal=3)

    pred = dkiM.predict(dkiF.model_params, S0=100)
    assert_array_almost_equal(pred, DWIsim_all_taylor * 100, decimal=3)

    # Check predict of KurtosisMicrostruturalFit
    pred = dkiF.predict(gtab_2s, S0=100)
    assert_array_almost_equal(pred, DWIsim_all_taylor * 100, decimal=3)


def _help_test_awf_only(dkimicrofit, string):
    exec(string)


def test_dki_micro_awf_only():
    dkiM = dki_micro.KurtosisMicrostructureModel(gtab_2s)
    dkiF = dkiM.fit(DWIsim, awf_only=True)
    awf = dkiF.awf
    assert_almost_equal(awf, FIE, decimal=3)

    # assert_raises(dkiF.hindered_evals)
    assert_raises(ValueError, _help_test_awf_only, dkiF, "dkimicrofit.hindered_evals")
    assert_raises(ValueError, _help_test_awf_only, dkiF, "dkimicrofit.restricted_evals")
    assert_raises(
        ValueError, _help_test_awf_only, dkiF, "dkimicrofit.axonal_diffusivity"
    )
    assert_raises(ValueError, _help_test_awf_only, dkiF, "dkimicrofit.hindered_ad")
    assert_raises(ValueError, _help_test_awf_only, dkiF, "dkimicrofit.hindered_rd")
    assert_raises(ValueError, _help_test_awf_only, dkiF, "dkimicrofit.tortuosity")


def additional_tortuosity_tests():
    # Test tortuosity when rd is zero
    # single voxel
    t = dki_micro.tortuosity(1.7e-3, 0.0)
    assert_almost_equal(t, 0.0)

    # multi-voxel
    RDEc = RDE.copy()
    Torc = Tor.copy()
    RDEc[1, 1, 1] = 0.0
    Torc[1, 1, 1] = 0.0
    t = dki_micro.tortuosity(ADE, RDEc)
    assert_almost_equal(Torc, t)



import numpy as np

from dipy.core.geometry import vec2vec_rotmat
from dipy.core.gradients import gradient_table
from dipy.data import get_fnames
import dipy.sims.voxel as vox

# import scipy.stats as stats
from dipy.sims.voxel import diffusion_evals, single_tensor
from dipy.testing.decorators import warning_for_keywords


@warning_for_keywords()
def add_noise(vol, *, snr=1.0, S0=None, noise_type="rician", rng=None):
    """Add noise of specified distribution to a 4D array.

    Parameters
    ----------
    vol : array, shape (X,Y,Z,W)
        Diffusion measurements in `W` directions at each ``(X, Y, Z)`` voxel
        position.
    snr : float, optional
        The desired signal-to-noise ratio.  (See notes below.)
    S0 : float, optional
        Reference signal for specifying `snr`.
    noise_type : string, optional
        The distribution of noise added. Can be either 'gaussian' for Gaussian
        distributed noise, 'rician' for Rice-distributed noise (default) or
        'rayleigh' for a Rayleigh distribution.
    rng : numpy.random.Generator class, optional
        Numpy's random generator for setting seed values when needed.

    Returns
    -------
    vol : array, same shape as vol
        Volume with added noise.

    Notes
    -----
    SNR is defined here, following :footcite:p:`Descoteaux2007`, as
    ``S0 / sigma``, where ``sigma`` is the standard deviation of the two
    Gaussian distributions forming the real and imaginary components of the
    Rician noise distribution (see :footcite:p:`Gudbjartsson1995`).

    References
    ----------
    .. footbibliography::

    Examples
    --------
    >>> signal = np.arange(800).reshape(2, 2, 2, 100)
    >>> signal_w_noise = add_noise(signal, snr=10, noise_type='rician')

    """
    orig_shape = vol.shape
    vol_flat = np.reshape(vol.copy(), (-1, vol.shape[-1]))

    if S0 is None:
        S0 = np.max(vol)

    for vox_idx, signal in enumerate(vol_flat):
        vol_flat[vox_idx] = vox.add_noise(
            signal, snr=snr, S0=S0, noise_type=noise_type, rng=rng
        )

    return np.reshape(vol_flat, orig_shape)


def diff2eigenvectors(dx, dy, dz):
    """numerical derivatives 2 eigenvectors"""
    basis = np.eye(3)
    u = np.array([dx, dy, dz])
    u = u / np.linalg.norm(u)
    R = vec2vec_rotmat(basis[:, 0], u)
    eig0 = u
    eig1 = np.dot(R, basis[:, 1])
    eig2 = np.dot(R, basis[:, 2])
    eigs = np.zeros((3, 3))
    eigs[:, 0] = eig0
    eigs[:, 1] = eig1
    eigs[:, 2] = eig2
    return eigs, R


@warning_for_keywords()
def orbital_phantom(
    *,
    gtab=None,
    evals=diffusion_evals,
    func=None,
    t=None,
    datashape=(64, 64, 64, 65),
    origin=(32, 32, 32),
    scale=(25, 25, 25),
    angles=None,
    radii=None,
    S0=100.0,
    snr=None,
    rng=None,
):
    """Create a phantom based on a 3-D orbit ``f(t) -> (x,y,z)``.

    Parameters
    ----------
    gtab : GradientTable
        Gradient table of measurement directions.
    evals : array, shape (3,)
        Tensor eigenvalues.
    func : user defined function f(t)->(x,y,z)
        It could be desirable for ``-1=<x,y,z <=1``.
        If None creates a circular orbit.
    t : array, shape (K,)
        Represents time for the orbit. Default is
        ``np.linspace(0, 2 * np.pi, 1000)``.
    datashape : array, shape (X,Y,Z,W)
        Size of the output simulated data
    origin : tuple, shape (3,)
        Define the center for the volume
    scale : tuple, shape (3,)
        Scale the function before applying to the grid
    angles : array, shape (L,)
        Density angle points, always perpendicular to the first eigen vector
        Default np.linspace(0, 2 * np.pi, 32).
    radii : array, shape (M,)
        Thickness radii.  Default ``np.linspace(0.2, 2, 6)``.
        angles and radii define the total thickness options
    S0 : double, optional
        Maximum simulated signal. Default 100.
    snr : float, optional
        The signal to noise ratio set to apply Rician noise to the data.
        Default is to not add noise at all.
    rng : numpy.random.Generator class, optional
        Numpy's random generator for setting seed values when needed.
        Default is None.

    Returns
    -------
    data : array, shape (datashape)

    See Also
    --------
    add_noise

    Examples
    --------

    >>> def f(t):
    ...    x = np.sin(t)
    ...    y = np.cos(t)
    ...    z = np.linspace(-1, 1, len(x))
    ...    return x, y, z

    >>> data = orbital_phantom(func=f)

    """

    if t is None:
        t = np.linspace(0, 2 * np.pi, 1000)

    if angles is None:
        angles = np.linspace(0, 2 * np.pi, 32)

    if radii is None:
        radii = np.linspace(0.2, 2, 6)

    if gtab is None:
        fimg, fbvals, fbvecs = get_fnames(name="small_64D")
        gtab = gradient_table(fbvals, bvecs=fbvecs)

    if func is None:
        x = np.sin(t)
        y = np.cos(t)
        z = np.zeros(t.shape)
    else:
        x, y, z = func(t)

    dx = np.diff(x)
    dy = np.diff(y)
    dz = np.diff(z)

    x = scale[0] * x + origin[0]
    y = scale[1] * y + origin[1]
    z = scale[2] * z + origin[2]

    bx = np.zeros(len(angles))
    by = np.sin(angles)
    bz = np.cos(angles)

    # The entire volume is considered to be inside the brain.
    # Voxels without a fiber crossing through them are taken
    # to be isotropic with signal = S0.
    vol = np.zeros(datashape) + S0

    for i in range(len(dx)):
        evecs, R = diff2eigenvectors(dx[i], dy[i], dz[i])
        S = single_tensor(gtab, S0, evals=evals, evecs=evecs, snr=None)

        vol[int(x[i]), int(y[i]), int(z[i]), :] += S

        for r in radii:
            for j in range(len(angles)):
                rb = np.dot(R, np.array([bx[j], by[j], bz[j]]))

                ix = int(x[i] + r * rb[0])
                iy = int(y[i] + r * rb[1])
                iz = int(z[i] + r * rb[2])
                vol[ix, iy, iz] = vol[ix, iy, iz] + S

    vol = vol / np.max(vol, axis=-1)[..., np.newaxis]
    vol *= S0

    if snr is not None:
        vol = add_noise(vol, snr=snr, S0=S0, noise_type="rician", rng=rng)

    return vol


if __name__ == "__main__":
    # TODO: this can become a nice tutorial for generating phantoms

    def f(t):
        x = np.sin(t)
        y = np.cos(t)
        # z=np.zeros(t.shape)
        z = np.linspace(-1, 1, len(x))
        return x, y, z

    # helix
    vol = orbital_phantom(func=f)

    def f2(t):
        x = np.linspace(-1, 1, len(t))
        y = np.linspace(-1, 1, len(t))
        z = np.zeros(x.shape)
        return x, y, z

    # first direction
    vol2 = orbital_phantom(func=f2)

    def f3(t):
        x = np.linspace(-1, 1, len(t))
        y = -np.linspace(-1, 1, len(t))
        z = np.zeros(x.shape)
        return x, y, z

    # second direction
    vol3 = orbital_phantom(func=f3)
    # double crossing
    vol23 = vol2 + vol3

    # """
    def f4(t):
        x = np.zeros(t.shape)
        y = np.zeros(t.shape)
        z = np.linspace(-1, 1, len(t))
        return x, y, z

    # triple crossing
    vol4 = orbital_phantom(func=f4)
    vol234 = vol23 + vol4

    # unknown function
    # voln = add_rician_noise(vol234)

    # """

    # from dipy.viz import window, actor
    # scene = window.Scene()
    # scene.add(actor.volume(vol234[...,0]))
    # window.show(scene)
    # vol234n=add_rician_noise(vol234,20)



import glob
import os
import shutil
import time

from sphinx_gallery.scrapers import figure_rst


class ImageFileScraper:
    def __init__(self):
        """Scrape image files that are already present in current folder."""
        self.embedded_images = {}
        self.start_time = time.time()

    def __call__(self, block, block_vars, gallery_conf):
        # Find all image files in the current directory.
        path_example = os.path.dirname(block_vars['src_file'])
        image_files = _find_images(path_example)
        # Iterate through files, copy them to the SG output directory
        image_names = []
        image_path_iterator = block_vars['image_path_iterator']
        for path_orig in image_files:
            # If we already know about this image and it hasn't been modified
            # since starting, then skip it
            mod_time = os.stat(path_orig).st_mtime
            already_embedded = (path_orig in self.embedded_images and
                                mod_time <= self.embedded_images[path_orig])
            existed_before_build = mod_time <= self.start_time
            if already_embedded or existed_before_build:
                continue
            # Else, we assume the image has just been modified and is displayed
            path_new = next(image_path_iterator)
            self.embedded_images[path_orig] = mod_time
            image_names.append(path_new)
            shutil.copyfile(path_orig, path_new)

        if len(image_names) == 0:
            return ''
        else:
            return figure_rst(image_names, gallery_conf['src_dir'])


def _find_images(path, image_extensions=['jpg', 'jpeg', 'png', 'gif']):
    """Find all unique image paths for a set of extensions."""
    image_files = set()
    for ext in image_extensions:
        this_ext_files = set(glob.glob(os.path.join(path, '*.'+ext)))
        image_files = image_files.union(this_ext_files)
    return image_files